# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod

from ..messages.message import Message


class Publisher(ABC):
    """An Abstract Base Class (ABC) for publishing Messages

    When updating to Python > 3.7 this should be converted into a
    typing.Protocol
    """

    @abstractmethod
    def publish(self, message: Message) -> None:
        raise NotImplementedError()



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Module to store Notus Scanner configuration settings
"""

import logging
import os
from pathlib import Path
from typing import Any, Dict

import tomli

from notus.scanner.errors import ConfigFileError

logger = logging.getLogger(__name__)

DEFAULT_PRODUCTS_DIRECTORY = "/var/lib/openvas/plugins/notus/products"
DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_MQTT_BROKER_ADDRESS = "localhost"
DEFAULT_MQTT_BROKER_PORT = 1883
DEFAULT_PID_FILE = "/run/notus-scanner/notus-scanner.pid"

_CONFIG = (
    (
        "products-directory",
        "NOTUS_SCANNER_PRODUCTS_DIRECTORY",
        DEFAULT_PRODUCTS_DIRECTORY,
    ),
    ("log-file", "NOTUS_SCANNER_LOG_FILE", None),
    ("log-level", "NOTUS_SCANNER_LOG_LEVEL", DEFAULT_LOG_LEVEL),
    (
        "mqtt-broker-address",
        "NOTUS_SCANNER_MQTT_BROKER_ADDRESS",
        DEFAULT_MQTT_BROKER_ADDRESS,
    ),
    (
        "mqtt-broker-port",
        "NOTUS_SCANNER_MQTT_BROKER_PORT",
        DEFAULT_MQTT_BROKER_PORT,
    ),
    ("pid-file", "NOTUS_SCANNER_PID_FILE", DEFAULT_PID_FILE),
    (
        "disable-hashsum-verification",
        "NOTUS_DISABLE_HASHSUM_VERIFICATION",
        False,
    ),
)


class Config:
    def __init__(self) -> None:
        self._config: Dict[str, Any] = {}

    def load(self, filepath: Path) -> None:
        try:
            content = filepath.read_text(encoding="utf-8")
            config_data = tomli.loads(content)
        except IOError as e:
            raise ConfigFileError(
                f"Can't load config file {filepath.absolute()}. Error was {e}."
            ) from e
        except tomli.TOMLDecodeError as e:
            raise ConfigFileError(
                f"Can't load config file. {filepath.absolute()} is not a valid "
                "TOML file."
            ) from e

        self._config = config_data.get("notus-scanner", {})

    def values(self) -> Dict[str, Any]:
        values = {}

        for config_key, env_key, default in _CONFIG:
            if env_key in os.environ:
                values[config_key] = os.environ.get(env_key)
            elif config_key in self._config:
                values[config_key] = self._config.get(config_key)
            else:
                values[config_key] = default

        return values



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Iterable, List, Set, Optional

from notus.scanner.models.packages import package_class_by_type

from .errors import AdvisoriesLoadingError
from .loader import AdvisoriesLoader
from .messages.message import Message
from .messages.result import ResultMessage
from .messages.start import ScanStartMessage
from .messages.status import ScanStatus, ScanStatusMessage
from .messaging.publisher import Publisher
from .models.packages.package import Package, PackageAdvisories, PackageAdvisory
from .models.vulnerability import Vulnerabilities, Vulnerability

logger = logging.getLogger(__name__)


class NotusScanner:
    def __init__(
        self,
        loader: AdvisoriesLoader,
        publisher: Publisher,
    ):
        self._loader = loader
        self._publisher = publisher

    def _publish(self, message: Message):
        """Try to publish a message

        Ensures that a failure during publishing doesn't stop the scan or daemon
        """
        try:
            self._publisher.publish(message)
        except Exception as e:  # pylint: disable=broad-except
            logger.error(
                "An error occurred while publishing a %s message. Reason %s",
                type(message),
                str(e),
            )

    def _finish_host(self, scan_id: str, host_ip: str):
        """Send a message to the broker to inform a host is done."""

        scan_status_message = ScanStatusMessage(
            scan_id=scan_id, host_ip=host_ip, status=ScanStatus.FINISHED
        )
        self._publish(scan_status_message)

    def _start_host(self, scan_id: str, host_ip: str):
        """Send a message to the broker to inform a host scan has started."""
        scan_status_message = ScanStatusMessage(
            scan_id=scan_id, host_ip=host_ip, status=ScanStatus.RUNNING
        )
        self._publish(scan_status_message)

    def _publish_results(
        self,
        scan_id: str,
        host_ip: str,
        host_name: str,
        vulnerabilities: Vulnerabilities,
    ) -> None:
        for oid, vulnerability in vulnerabilities.get().items():
            report = ""
            fixed_packages: List[PackageAdvisory]
            for package, fixed_packages in vulnerability.get().items():
                fixed_package = fixed_packages.pop(0)
                report = (
                    report + f"\n{'Vulnerable package:':<22}{package.name}\n"
                    f"{'Installed version:':<22}{package.full_name}\n"
                    f"{'Fixed version:':<20}{fixed_package.symbol:>2}"
                    f"{fixed_package.package.full_name}\n"
                )
                for fixed_package in fixed_packages:
                    report = (
                        report + f"{'':<20}{fixed_package.symbol:>2}"
                        f"{fixed_package.package.full_name}\n"
                    )

            message = ResultMessage(
                scan_id=scan_id,
                host_ip=host_ip,
                host_name=host_name,
                oid=oid,
                value=report,
            )
            self._publish(message)

    @staticmethod
    def _check_package(
        package: Package, package_advisory_list: Set[PackageAdvisory]
    ) -> Optional[Vulnerability]:

        vul = Vulnerability()
        for package_advisory in package_advisory_list:
            logger.debug(
                "%s verify package %s %s %s",
                package_advisory.oid,
                package,
                package_advisory.symbol,
                package_advisory.package,
            )
            is_vulnerable = package_advisory.is_vulnerable(package)
            if is_vulnerable is None:
                continue
            elif not is_vulnerable:
                return

            vul.add(package, package_advisory)

        return vul

    def _start_scan(
        self,
        installed_packages: Iterable[Package],
        package_advisories: PackageAdvisories,
    ) -> Vulnerabilities:

        vulnerabilities = Vulnerabilities()

        for package in installed_packages:
            package_advisory_oids = (
                package_advisories.get_package_advisories_for_package(package)
            )
            for oid, package_advisory_list in package_advisory_oids.items():

                vul = self._check_package(package, package_advisory_list)
                if vul and vul.vulnerability:
                    vulnerabilities.add(oid, vul)

        return vulnerabilities

    def run_scan(
        self,
        message: ScanStartMessage,
    ) -> None:
        """Handle the data necessary to start a scan,
        received via mqtt and run the scan."""

        # Check if all necessary information to run a scan are given
        if not message:
            logger.error(
                "Unable to start scan for %s: The message seems to be empty",
                message.host_ip,
            )
            return
        if not message.os_release:
            logger.error(
                "Unable to start scan for %s: The field os_release is empty",
                message.host_ip,
            )
            return
        if not message.package_list:
            logger.error(
                "Unable to start scan for %s: The field package_list is empty",
                message.host_ip,
            )
            return

        # Get advisory information from disk
        try:
            package_advisories = self._loader.load_package_advisories(
                message.os_release
            )
        except AdvisoriesLoadingError as e:
            logger.error("Unable to load package advisories. Error was %s", e)
            return

        if not package_advisories:
            # Probably a wrong or not supported OS-release
            logger.error(
                "Unable to start scan for %s: No advisories for OS-release %s"
                " found. Check if the OS-release is correct and the"
                " corresponding advisories are given.",
                message.host_ip,
                message.os_release,
            )
            return

        logger.debug(
            "Loaded advisories for %i packages", len(package_advisories)
        )

        # Determine package type
        package_type = package_advisories.package_type

        package_class = package_class_by_type(package_type)
        if not package_class:
            logger.error(
                "Unable to start scan for %s: No package implementation for "
                "OS-release %s found. Check if the OS-release is correct.",
                message.host_ip,
                message.os_release,
            )
            return

        may_installed = [
            package_class.from_full_name(name) for name in message.package_list
        ]
        # a package in may_installed can only be None when .from_full_name fails
        # they both log a warning when they're unable to parse that hence it
        # is safe to silently remove them
        installed_packages: Iterable[Package] = (
            package for package in may_installed if package is not None
        )

        self._start_host(message.scan_id, message.host_ip)

        logger.info(
            "Start to identify vulnerable packages for %s (%s)",
            message.host_ip,
            message.host_name,
        )
        try:
            vulnerabilities = self._start_scan(
                installed_packages=installed_packages,
                package_advisories=package_advisories,
            )
            self._publish_results(
                message.scan_id,
                message.host_ip,
                message.host_name,
                vulnerabilities,
            )

            logger.info(
                "Total number of vulnerable packages -> %d",
                len(vulnerabilities),
            )

            self._finish_host(message.scan_id, message.host_ip)

        except AdvisoriesLoadingError as e:
            logger.error(
                "Scan for %s %s with %s could not be started. Error was %s",
                message.host_ip,
                message.host_name or "",
                message.os_release,
                e,
            )



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Module for parsing and comparing Slackware packages
"""

import logging
import re
from dataclasses import dataclass

from packaging.version import parse

from .package import Package, PackageComparison

_slack_compile = re.compile(r"(..*)-(..*)-(..*)-(\d)(?:_slack(..*))?")
_slack_compile_version = re.compile(r"(..*)-(..*)-(\d)(?:_slack(..*))?")


logger = logging.getLogger(__name__)


@dataclass
class SlackPackage(Package):
    """Represents a Slackware based package"""

    build: str
    target: str
    version: str
    arch: str

    __hash__ = Package.__hash__

    def compare(self, other: "SlackPackage") -> PackageComparison:
        if self.name != other.name:
            return PackageComparison.NOT_COMPARABLE

        if self.arch != other.arch:
            return PackageComparison.NOT_COMPARABLE

        if self.full_version == other.full_version:
            return PackageComparison.EQUAL

        comp = self.version_compare(self.version, other.version)
        if comp != PackageComparison.EQUAL:
            return comp

        a_target = parse(self.target)
        b_target = parse(other.target)

        if a_target and b_target and a_target != b_target:
            return (
                PackageComparison.A_NEWER
                if a_target > b_target
                else PackageComparison.B_NEWER
            )
        comp = self.version_compare(self.target, other.target)
        if comp != PackageComparison.EQUAL:
            return comp

        return self.version_compare(self.build, other.build)

    @staticmethod
    def from_full_name(full_name: str):
        if not full_name:
            return None

        full_name = full_name.strip()
        match = _slack_compile.match(full_name)
        if match:
            name, version, arch, build, target = match.groups()
        else:
            logger.warning(
                "The slack package %s could not be parsed", full_name
            )
            return None

        full_version = f"{version}-{arch}-{build}"

        if target:
            full_version = f"{full_version}_slack{target}"

        return SlackPackage(
            name=name,
            arch=arch,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
            build=build,
            target=target or "",
            version=version,
        )

    @staticmethod
    def from_name_and_full_version(name: str, full_version: str):
        if not name or not full_version:
            return None
        name = name.strip()
        full_version = full_version.strip()
        match = _slack_compile_version.match(full_version)
        if match:
            version, arch, build, target = match.groups()
        else:
            logger.warning(
                "The slack package %s could not be parsed",
                f"{name}-{full_version}",
            )
            return None

        return SlackPackage(
            name=name,
            arch=arch,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
            build=build,
            target=target or "",
            version=version,
        )



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase

from notus.scanner.models.packages.deb import DEBPackage


class DEBPackageTestCase(TestCase):
    def test_compare_gt(self):
        """packages should be comparable"""
        package1 = DEBPackage(
            name="foo-bar",
            epoch="1",
            upstream_version="1.2.3",
            debian_revision="4",
            full_name="foo-bar-1:1.2.3-4",
            full_version="1:1.2.3-4",
        )
        package2 = DEBPackage(
            name="foo-bar",
            epoch="1",
            upstream_version="1.2.4",
            debian_revision="4",
            full_name="foo-bar-1:1.2.4-4",
            full_version="1:1.2.4-4",
        )
        self.assertGreater(package2, package1)

        package2 = DEBPackage(
            name="foo-bar",
            epoch="1",
            upstream_version="1.2.3",
            debian_revision="5",
            full_name="foo-bar-1:1.2.3-5",
            full_version="1:1.2.3-5",
        )
        self.assertGreater(package2, package1)

    def test_compare_gt_different_name(self):
        """different packagtes should not be comparable"""
        package1 = DEBPackage(
            name="foo",
            epoch="1",
            upstream_version="1.2.3",
            debian_revision="4",
            full_name="foo-1:1.2.3-4",
            full_version="1:1.2.3-4",
        )
        package2 = DEBPackage(
            name="bar",
            epoch="1",
            upstream_version="1.2.3",
            debian_revision="4",
            full_name="bar-1:1.2.3-4",
            full_version="1:1.2.3-4",
        )
        self.assertFalse(package2 > package1)
        self.assertFalse(package1 > package2)

    def test_compare_less(self):
        """packages should be comparable"""
        package1 = DEBPackage(
            name="foo-bar",
            epoch="1",
            upstream_version="1.2.3",
            debian_revision="4",
            full_name="foo-bar-1:1.2.3-4",
            full_version="1:1.2.3-4",
        )
        package2 = DEBPackage(
            name="foo-bar",
            epoch="1",
            upstream_version="1.2.4",
            debian_revision="4",
            full_name="foo-bar-1:1.2.4-4",
            full_version="1:1.2.4-4",
        )
        self.assertLess(package1, package2)

        package2 = DEBPackage(
            name="foo-bar",
            epoch="1",
            upstream_version="1.2.3",
            debian_revision="5",
            full_name="foo-bar-1.2.3-5:1",
            full_version="1:1.2.3-5",
        )
        self.assertLess(package1, package2)

        package2 = DEBPackage(
            name="foo-bar",
            epoch="1",
            upstream_version="1.2.3~rc",
            debian_revision="4",
            full_name="foo-bar-1:1.2.3~rc-4",
            full_version="1:1.2.3~rc-4",
        )
        self.assertLess(package2, package1)

    def test_compare_equal(self):
        """packages with the same data should be equal"""
        package1 = DEBPackage(
            name="foo-bar",
            epoch="1",
            upstream_version="1.2.3",
            debian_revision="4",
            full_name="foo-bar-1:1.2.3-4",
            full_version="1:1.2.3-4",
        )
        package2 = DEBPackage(
            name="foo-bar",
            epoch="1",
            upstream_version="1.2.3",
            debian_revision="4",
            full_name="foo-bar-1:1.2.3-4",
            full_version="1:1.2.3-4",
        )

        self.assertEqual(package1, package2)

    def test_from_full_name(self):
        """it should be possible to create packages via the full name"""
        self.assertIsNone(DEBPackage.from_full_name(None))

        package = DEBPackage.from_full_name("mesa-libgbm-2:11.2.2-2.20160614")
        self.assertEqual(package.name, "mesa-libgbm")
        self.assertEqual(package.epoch, "2")
        self.assertEqual(package.upstream_version, "11.2.2")
        self.assertEqual(package.debian_revision, "2.20160614")
        self.assertEqual(package.full_name, "mesa-libgbm-2:11.2.2-2.20160614")
        self.assertEqual(package.full_version, "2:11.2.2-2.20160614")

        package = DEBPackage.from_full_name("keyutils-1.5.8-3")
        self.assertEqual(package.name, "keyutils")
        self.assertEqual(package.epoch, "0")
        self.assertEqual(package.upstream_version, "1.5.8")
        self.assertEqual(package.debian_revision, "3")
        self.assertEqual(package.full_name, "keyutils-1.5.8-3")
        self.assertEqual(package.full_version, "1.5.8-3")

        package = DEBPackage.from_full_name("httpd-manual-1:2.4.6-45.0.1.4.h10")
        self.assertEqual(package.name, "httpd-manual")
        self.assertEqual(package.epoch, "1")
        self.assertEqual(package.upstream_version, "2.4.6")
        self.assertEqual(package.debian_revision, "45.0.1.4.h10")
        self.assertEqual(package.full_name, "httpd-manual-1:2.4.6-45.0.1.4.h10")

        package = DEBPackage.from_full_name("libzstd1-1.3.8+dfsg-3+deb10u2")
        self.assertEqual(package.name, "libzstd1")
        self.assertEqual(package.epoch, "0")
        self.assertEqual(package.upstream_version, "1.3.8+dfsg")
        self.assertEqual(package.debian_revision, "3+deb10u2")
        self.assertEqual(package.full_name, "libzstd1-1.3.8+dfsg-3+deb10u2")

        package = DEBPackage.from_full_name(
            "xserver-xorg-video-intel-2:2.99.917+git20180925-2"
        )
        self.assertEqual(package.name, "xserver-xorg-video-intel")
        self.assertEqual(package.epoch, "2")
        self.assertEqual(package.upstream_version, "2.99.917+git20180925")
        self.assertEqual(package.debian_revision, "2")
        self.assertEqual(
            package.full_name,
            "xserver-xorg-video-intel-2:2.99.917+git20180925-2",
        )

        package = DEBPackage.from_full_name("ucf-3.0038+nmu1")
        self.assertEqual(package.name, "ucf")
        self.assertEqual(package.epoch, "0")
        self.assertEqual(package.upstream_version, "3.0038+nmu1")
        self.assertEqual(package.debian_revision, "")
        self.assertEqual(package.full_name, "ucf-3.0038+nmu1")

        package = DEBPackage.from_full_name("apport-symptoms-020")
        self.assertEqual(package.name, "apport-symptoms")
        self.assertEqual(package.epoch, "0")
        self.assertEqual(package.upstream_version, "020")
        self.assertEqual(package.debian_revision, "")
        self.assertEqual(package.full_name, "apport-symptoms-020")

    def test_from_name_and_full_version(self):
        """it should be possible to create packages from name and full
        version"""
        self.assertIsNone(DEBPackage.from_name_and_full_version(None, None))

        package = DEBPackage.from_name_and_full_version(
            "mesa-libgbm", "2:11.2.2-2.20160614"
        )
        self.assertEqual(package.name, "mesa-libgbm")
        self.assertEqual(package.epoch, "2")
        self.assertEqual(package.upstream_version, "11.2.2")
        self.assertEqual(package.debian_revision, "2.20160614")
        self.assertEqual(package.full_name, "mesa-libgbm-2:11.2.2-2.20160614")
        self.assertEqual(package.full_version, "2:11.2.2-2.20160614")

        package = DEBPackage.from_name_and_full_version(
            "mesa-libgbm", "2:11.2.2"
        )
        self.assertEqual(package.name, "mesa-libgbm")
        self.assertEqual(package.epoch, "2")
        self.assertEqual(package.upstream_version, "11.2.2")
        self.assertEqual(package.debian_revision, "")
        self.assertEqual(package.full_name, "mesa-libgbm-2:11.2.2")
        self.assertEqual(package.full_version, "2:11.2.2")

        package = DEBPackage.from_name_and_full_version("mesa-libgbm", "11.2.2")
        self.assertEqual(package.name, "mesa-libgbm")
        self.assertEqual(package.epoch, "0")
        self.assertEqual(package.upstream_version, "11.2.2")
        self.assertEqual(package.debian_revision, "")
        self.assertEqual(package.full_name, "mesa-libgbm-11.2.2")
        self.assertEqual(package.full_version, "11.2.2")

        package = DEBPackage.from_name_and_full_version(
            "mesa-libgbm", "11.2.2-2.20160614"
        )
        self.assertEqual(package.name, "mesa-libgbm")
        self.assertEqual(package.epoch, "0")
        self.assertEqual(package.upstream_version, "11.2.2")
        self.assertEqual(package.debian_revision, "2.20160614")
        self.assertEqual(package.full_name, "mesa-libgbm-11.2.2-2.20160614")
        self.assertEqual(package.full_version, "11.2.2-2.20160614")



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from ..errors import MessageParsingError
from .message import Message, MessageType


class ScanStartMessage(Message):
    message_type: MessageType = MessageType.SCAN_START
    topic = "scanner/package/cmd/notus"

    scan_id: str
    host_ip: str
    host_name: str
    os_release: str
    package_list: List[str]

    def __init__(
        self,
        *,
        scan_id: str,
        host_ip: str,
        host_name: str,
        os_release: str,
        package_list: List[str],
        message_id: Optional[UUID] = None,
        group_id: Optional[str] = None,
        created: Optional[datetime] = None,
    ):
        super().__init__(
            message_id=message_id, group_id=group_id, created=created
        )
        self.scan_id = scan_id
        self.host_ip = host_ip
        self.host_name = host_name
        self.os_release = os_release
        self.package_list = package_list if package_list is not None else []

    def serialize(self) -> Dict[str, Union[int, str, List[str]]]:
        message = super().serialize()
        message.update(
            {
                "scan_id": self.scan_id,
                "host_ip": self.host_ip,
                "host_name": self.host_name,
                "os_release": self.os_release,
                "package_list": self.package_list,
            }
        )
        return message

    @classmethod
    def _parse(cls, data: Dict[str, Union[int, str]]) -> Dict[str, Any]:
        kwargs = super()._parse(data)

        package_list = data.get("package_list")
        if not isinstance(package_list, list):
            raise MessageParsingError("package_list must contain a list")

        kwargs.update(
            {
                "scan_id": data.get("scan_id"),
                "host_ip": data.get("host_ip"),
                "host_name": data.get("host_name"),
                "os_release": data.get("os_release"),
                "package_list": package_list,
            }
        )
        return kwargs



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
import re
from abc import abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, Dict, Optional, Set

from ...errors import PackageError

logger = logging.getLogger(__name__)


version_component_re = re.compile(r"(\d+ | .)", re.X)


class PackageType(Enum):
    RPM = "rpm"
    DEB = "deb"
    EBUILD = "ebuild"
    SLACK = "slack"


class PackageComparison(Enum):
    EQUAL = 0  # a and b are equal
    A_NEWER = 1  # a is newer than b
    B_NEWER = 2  # b is newer than a
    # a and b are not comparable. e.g. different architectures
    NOT_COMPARABLE = 3


@dataclass
class Package:
    "Base class for different Package types"
    name: str
    full_name: str
    full_version: str

    def __guard_cmp(self, other: Any):
        if not isinstance(other, type(self)):
            raise PackageError(f"Can't compare {self!r} to {other!r}.")

    def __gt__(self, other: Any) -> bool:
        self.__guard_cmp(other)
        return self.compare(other) == PackageComparison.A_NEWER

    def __lt__(self, other: Any) -> bool:
        self.__guard_cmp(other)
        return self.compare(other) == PackageComparison.B_NEWER

    def __eq__(self, other: Any) -> bool:
        self.__guard_cmp(other)
        return self.compare(other) == PackageComparison.EQUAL

    def __ge__(self, other: Any) -> bool:
        return self.__gt__(other) or self.__eq__(other)

    def __le__(self, other: Any) -> bool:
        return self.__lt__(other) or self.__eq__(other)

    def __hash__(self) -> int:
        # allow to hash the package
        # the full name identifies the package
        return hash(self.full_name)

    @staticmethod
    def version_compare(version_a: str, version_b: str) -> PackageComparison:
        """Compares two versions"""
        if version_a == version_b:
            return PackageComparison.EQUAL

        a_parts = version_component_re.split(version_a)
        b_parts = version_component_re.split(version_b)

        for i in range(max(len(a_parts), len(b_parts))):
            if i < len(a_parts):
                a_part = a_parts[i]
            else:
                return (
                    PackageComparison.B_NEWER
                    if b_parts[i] != "~"
                    else PackageComparison.A_NEWER
                )
            if i < len(b_parts):
                b_part = b_parts[i]
            else:
                return (
                    PackageComparison.A_NEWER
                    if a_part != "~"
                    else PackageComparison.B_NEWER
                )
            if a_part == b_part:
                continue

            if a_part.isnumeric() and b_part.isnumeric():
                return (
                    PackageComparison.A_NEWER
                    if int(a_part) > int(b_part)
                    else PackageComparison.B_NEWER
                )
            if a_part.isnumeric() or b_part.isnumeric():
                return (
                    PackageComparison.A_NEWER
                    if a_part.isnumeric()
                    else PackageComparison.B_NEWER
                )

            if a_part.isalpha() and b_part.isalpha():
                return (
                    PackageComparison.A_NEWER
                    if a_part.lower() > b_part.lower()
                    else PackageComparison.B_NEWER
                )
            if a_part.isalpha():
                return (
                    PackageComparison.A_NEWER
                    if b_part == "~"
                    else PackageComparison.B_NEWER
                )
            if b_part.isalpha():
                return (
                    PackageComparison.B_NEWER
                    if a_part == "~"
                    else PackageComparison.A_NEWER
                )

            return (
                PackageComparison.A_NEWER
                if a_part != "~" and a_part > b_part or b_part == "~"
                else PackageComparison.B_NEWER
            )
        return PackageComparison.EQUAL

    @abstractmethod
    def compare(self, other: Any) -> PackageComparison:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def from_full_name(full_name: str):
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def from_name_and_full_version(name: str, full_version: str):
        raise NotImplementedError()


@dataclass(frozen=True, unsafe_hash=True)
class PackageAdvisory:
    """Connects a package with an advisory"""

    package: Package
    oid: str
    symbol: str
    # returns None when not comparable otherwise true or false
    is_vulnerable: Callable[[Package], Optional[bool]] = field(
        compare=False, hash=False
    )


@dataclass(frozen=True)
class PackageAdvisories:
    """Container for mapping a package name to a set of advisories for this
    package"""

    package_type: PackageType
    advisories: Dict[str, Dict[str, Set[PackageAdvisory]]] = field(
        default_factory=dict
    )

    is_comparable = (
        lambda a, b: a.compare(b) != PackageComparison.NOT_COMPARABLE
    )

    comparison_map = {
        ">=": lambda a, b: a > b
        if PackageAdvisories.is_comparable(a, b)
        else None,
        "<=": lambda a, b: a < b
        if PackageAdvisories.is_comparable(a, b)
        else None,
        "=": lambda a, b: a != b
        if PackageAdvisories.is_comparable(a, b)
        else None,
        "<": lambda a, b: a <= b
        if PackageAdvisories.is_comparable(a, b)
        else None,
        ">": lambda a, b: a >= b
        if PackageAdvisories.is_comparable(a, b)
        else None,
    }

    def get_package_advisories_for_package(
        self, package: Package
    ) -> Dict[str, Set[PackageAdvisory]]:
        return self.advisories.get(package.name) or dict()

    def add_advisory_for_package(
        self,
        package: Package,
        advisory: str,
        verifier: Optional[str],
    ) -> None:
        if verifier not in self.comparison_map:
            verifier = ">="
        advisories = self.get_package_advisories_for_package(package)
        is_vulnerable = lambda other: self.comparison_map[verifier](
            package, other
        )

        if not advisory in advisories:
            advisories[advisory] = set()

        advisories[advisory].add(
            PackageAdvisory(package, advisory, verifier, is_vulnerable)
        )
        self.advisories[package.name] = advisories

    def __len__(self) -> int:
        return len(self.advisories)



# -*- coding: utf-8 -*-
# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import logging
import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, Optional

from gnupg import GPG

OPENVAS_GPG_HOME = "/etc/openvas/gnupg"


logger = logging.getLogger(__name__)


def __determine_default_gpg_home() -> Path:
    gos_default = Path(OPENVAS_GPG_HOME)
    if gos_default.exists():
        return gos_default
    user_default = Path.home() / ".gnupg"
    if not user_default.exists():
        logger.warning(
            "No GnuPG home found; "
            "please verify setup and set the GNUPGHOME variable if necessary"
        )
    return user_default


def __default_gpg_home() -> GPG:
    """
    __defaultGpgHome tries to load the variable 'GNUPGHOME' or to guess it
    """
    manual = os.getenv("GNUPGHOME")

    home = Path(manual) if manual else __determine_default_gpg_home()
    logger.debug("Using %s as GnuPG home.", home)
    return GPG(gnupghome=f"{home.absolute()}")


@dataclass
class ReloadConfiguration:
    hash_file: Path
    on_verification_failure: Callable[
        [Optional[Dict[str, str]]], Dict[str, str]
    ]
    gpg: Optional[GPG] = None
    cache: Optional[Dict[str, str]] = None
    fingerprint: str = ""


def reload_sha256sums(
    config: ReloadConfiguration,
) -> Callable[[], Dict[str, str]]:
    """
    reload_sha256sums reloads sha256sums if a threshold has been reached.
    """
    if not config.gpg:
        config.gpg = __default_gpg_home()

    def create_hash(file: Path) -> str:
        # we just use the hash to identify we have to reload the sha256sums
        # therefore a collision is not the end of the world and sha1 is more
        # than sufficient
        hasher = hashlib.sha1()
        with file.open(mode="rb") as f:
            for hash_file_bytes in iter(lambda: f.read(1024), b""):
                hasher.update(hash_file_bytes)
        return hasher.hexdigest()

    def internal_reload() -> Dict[str, str]:
        fingerprint = create_hash(config.hash_file)
        if not config.cache or config.fingerprint != fingerprint:
            config.fingerprint = fingerprint
            config.cache = gpg_sha256sums(config.hash_file, config.gpg)
        if not config.cache:
            return config.on_verification_failure(None)
        return config.cache

    return internal_reload


def gpg_sha256sums(
    hash_file: Path, gpg: Optional[GPG] = None
) -> Optional[Dict[str, str]]:
    """
    gpg_sha256sums verifies given hash_file with a asc file

    This functions assumes that the asc file is in the same directory as the
    hashfile and has the same name but with the suffix '.asc'
    """

    # when doing that via paramater list it is loading eagerly on import
    # which may fail on some systems
    if not gpg:
        gpg = __default_gpg_home()
    asc_path = hash_file.parent / f"{hash_file.name}.asc"
    with asc_path.open(mode="rb") as f:
        verified = gpg.verify_file(f, str(hash_file.absolute()))
        if not verified:
            return None
        result = {}
        with hash_file.open() as f:
            for line in f.readlines():
                hsum, fname = line.split("  ")
                # the second part can contain a newline
                # sometimes the hash sum got generated outside the current dir
                # and may contain leading paths.
                # Since we check against the filename we should normalize to
                # prevent false positives.
                result[hsum] = fname.split("/")[-1].strip()
        return result


class VerificationResult(Enum):
    INVALID_FILE = 0
    INVALID_HASH = 1
    INVALID_NAME = 2
    SUCCESS = 3


def create_verify(
    sha256sums: Callable[[], Dict[str, str]]
) -> Callable[[Path], VerificationResult]:
    """
    create_verify is returning a closure based on the sha256sums.

    This allows to load sha256sums and verify there instead of verifying and
    loading on each verification request.
    """

    def verify(advisory_path: Path) -> VerificationResult:
        s256h = hashlib.sha256()
        if not advisory_path.is_file():
            return VerificationResult.INVALID_FILE

        with advisory_path.open(mode="rb") as f:
            for hash_file_bytes in iter(lambda: f.read(1024), b""):
                s256h.update(hash_file_bytes)
        hash_sum = s256h.hexdigest()

        assumed_name = sha256sums().get(hash_sum)
        if not assumed_name:
            return VerificationResult.INVALID_HASH
        return (
            VerificationResult.SUCCESS
            if assumed_name == advisory_path.name
            else VerificationResult.INVALID_NAME
        )

    return verify



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
from pathlib import Path
from uuid import uuid4

from ..cli.parser import log_level
from ..config import DEFAULT_MQTT_BROKER_ADDRESS, DEFAULT_MQTT_BROKER_PORT
from ..messages.start import ScanStartMessage
from ..messaging.mqtt import MQTTClient, MQTTPublisher


def after_publish(client, _userdata, _mid):
    client.disconnect()


def main():
    parser = argparse.ArgumentParser(
        description="A test client to start generating scan results via the "
        "Notus Scanner"
    )
    parser.add_argument(
        "-b",
        "--mqtt-broker-address",
        type=str,
        required=True,
        default=DEFAULT_MQTT_BROKER_ADDRESS,
        help="Hostname or IP address of the MQTT broker.",
    )
    parser.add_argument(
        "-p",
        "--mqtt-broker-port",
        type=int,
        default=DEFAULT_MQTT_BROKER_PORT,
        help="Port of the MQTT broker. (default: %(default)s)",
    )
    parser.add_argument(
        "-s",
        "--scan-id",
        help="ID to use for the scan. If no scan ID is provided a random ID "
        "will be generated",
    )
    parser.add_argument(
        "--host-ip",
        required=True,
        help="IP address of the host to report in generated results",
    )
    parser.add_argument(
        "--host-name", help="Name of the host to report in generated results"
    )
    parser.add_argument(
        "--os-release",
        required=True,
        help="Name of the Operating System Release to scan",
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--packages",
        nargs="+",
        help="List of packages to compare for vulnerabilities",
    )
    group.add_argument(
        "--package-file",
        type=Path,
        help="Path to a file containing a list of packages to compare for "
        "vulnerabilities",
    )
    parser.add_argument(
        "-L",
        "--log-level",
        default="INFO",
        type=log_level,
        help="Wished level of logging (default: %(default)s)",
    )

    args = parser.parse_args()

    logging.basicConfig(level=args.log_level)

    scan_id = args.scan_id or str(uuid4())
    if args.packages:
        package_list = args.packages
    else:
        package_file_path: Path = args.package_file
        with package_file_path.open("r", encoding="utf-8") as f:
            package_list = f.readlines()

    print(f"Starting a scan with ID {scan_id}")

    start_scan_message = ScanStartMessage(
        scan_id=scan_id,
        host_ip=args.host_ip,
        host_name=args.host_name,
        os_release=args.os_release,
        package_list=package_list,
    )
    client = MQTTClient(
        mqtt_broker_address=args.mqtt_broker_address,
        mqtt_broker_port=args.mqtt_broker_port,
        client_id=f"notus-scan-start-{scan_id}",
    )
    client.on_publish = after_publish
    client.connect()
    client.loop()

    publisher = MQTTPublisher(client)
    publisher.publish(start_scan_message)

    client.loop_forever()


if __name__ == "__main__":
    main()



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
import logging
from functools import partial
from typing import Callable, Type

import paho.mqtt.client as mqtt

from ..errors import MessageParsingError
from ..messages.message import Message
from .publisher import Publisher
from .subscriber import Subscriber

logger = logging.getLogger(__name__)

NOTUS_MQTT_CLIENT_ID = "notus-scanner"

QOS_AT_LEAST_ONCE = 1


class MQTTClient(mqtt.Client):
    def __init__(
        self,
        mqtt_broker_address: str,
        mqtt_broker_port: int,
        client_id=NOTUS_MQTT_CLIENT_ID,
    ):
        self._mqtt_broker_address = mqtt_broker_address
        self._mqtt_broker_port = mqtt_broker_port

        super().__init__(client_id=client_id, protocol=mqtt.MQTTv5)

        self.enable_logger()

    def connect(
        self,
        host=None,
        port=None,
        keepalive=60,
        bind_address="",
        bind_port=0,
        clean_start=mqtt.MQTT_CLEAN_START_FIRST_ONLY,
        properties=None,
    ):
        if not host:
            host = self._mqtt_broker_address
        if not port:
            port = self._mqtt_broker_port

        return super().connect(
            host,
            port=port,
            keepalive=keepalive,
            bind_address=bind_address,
            bind_port=bind_port,
            clean_start=clean_start,
            properties=properties,
        )


class MQTTPublisher(Publisher):
    def __init__(self, client: MQTTClient):
        self._client = client

    def publish(self, message: Message) -> None:
        logger.debug("Publish message %s", message)
        self._client.publish(message.topic, str(message), qos=QOS_AT_LEAST_ONCE)


class MQTTSubscriber(Subscriber):
    def __init__(self, client: MQTTClient):
        self._client = client
        # Save the active subscriptions on subscribe() so we can resubscribe
        # after reconnect
        self._subscriptions: dict = {}

        self._client.on_connect = self.on_connect
        self._client.user_data_set(self._subscriptions)

    def subscribe(
        self, message_class: Type[Message], callback: Callable[[Message], None]
    ) -> None:
        func = partial(self._handle_message, message_class, callback)
        func.__name__ = callback.__name__

        logger.debug("Subscribing to topic %s", message_class.topic)

        self._client.subscribe(message_class.topic, qos=QOS_AT_LEAST_ONCE)
        self._client.message_callback_add(message_class.topic, func)

        self._subscriptions[message_class.topic] = func

    @staticmethod
    def on_connect(_client, _userdata, _flags, rc, _properties):
        if rc == 0:
            # If we previously had active subscription we subscribe to them
            # again because they got lost after a broker disconnect.
            # Userdata was set in __init__()
            if _userdata:
                for topic, func in _userdata.items():
                    _client.subscribe(topic, qos=QOS_AT_LEAST_ONCE)
                    _client.message_callback_add(topic, func)

    @staticmethod
    def _handle_message(
        message_class: Type[Message],
        callback: Callable[[Message], None],
        _client,
        _userdata,
        msg: mqtt.MQTTMessage,
    ) -> None:
        logger.debug("Incoming message for topic %s", msg.topic)

        try:
            # Load message from payload
            message = message_class.load(msg.payload)
        except json.JSONDecodeError:
            logger.error(
                "Got MQTT message in non-json format for topic %s.", msg.topic
            )
            logger.debug("Got: %s", msg.payload)
            return
        except MessageParsingError as e:
            logger.error(
                "Could not parse message for topic %s. Error was %s",
                msg.topic,
                e,
            )
            logger.debug("Got: %s", msg.payload)
            return

        callback(message)


class MQTTDaemon:
    """A class to start and stop the MQTT client"""

    def __init__(
        self,
        client: MQTTClient,
    ):
        self._client = client

        self._client.connect()

    def run(self):
        self._client.loop_forever()



# slightly adjusted from https://github.com/ihiji/version_utils
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import re
from dataclasses import dataclass

from .package import Package, PackageComparison

_rpm_compile_no_arch = re.compile("(.*)-([^-]+)-([^-]+)")
_rpm_compile = re.compile(r"(.*)-([^-]+)-([^-]+)\.([^-]+)")
_rpm_compile_version = re.compile(r"([^-]+)-([^-]+)\.([^-]+)")

logger = logging.getLogger(__name__)

exceptions = [
    "_fips",
    ".ksplice",
]


@dataclass
class RPMPackage(Package):
    """Represents a RPM package"""

    version: str
    release: str
    arch: str

    __hash__ = Package.__hash__

    def compare(self, other: "RPMPackage") -> PackageComparison:
        if self.name != other.name:
            return PackageComparison.NOT_COMPARABLE

        if self.arch != other.arch:
            return PackageComparison.NOT_COMPARABLE

        for e in exceptions:
            if (self.full_version.find(e) > -1) != (
                other.full_version.find(e) > -1
            ):
                return PackageComparison.NOT_COMPARABLE

        if self.full_version == other.full_version:
            return PackageComparison.EQUAL

        comp = self.version_compare(self.version, other.version)
        if comp != PackageComparison.EQUAL:
            return comp

        return self.version_compare(self.release, other.release)

    @staticmethod
    def from_full_name(full_name: str):
        if not full_name:
            return None

        full_name = full_name.strip()

        match = _rpm_compile.match(full_name)
        if match:
            name, version, release, arch = match.groups()
        else:
            match = _rpm_compile_no_arch.match(full_name)
            if match:
                name, version, release = match.groups()
                arch = ""
            else:
                logger.warning(
                    "The rpm package %s could not be parsed", full_name
                )
                return None

        return RPMPackage(
            name=name,
            version=version,
            release=release,
            arch=arch,
            full_name=full_name,
            full_version=f"{version}-{release}.{arch}",
        )

    @staticmethod
    def from_name_and_full_version(name: str, full_version: str):
        if not name or not full_version:
            return None

        name = name.strip()
        full_version = full_version.strip()

        version_match = _rpm_compile_version.match(full_version)
        if not version_match:
            return None
        version, release, arch = version_match.groups()

        return RPMPackage(
            name=name,
            version=version,
            release=release,
            arch=arch,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
        )



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timezone
from unittest import TestCase
from uuid import UUID

from notus.scanner.errors import MessageParsingError
from notus.scanner.messages.message import MessageType
from notus.scanner.messages.start import ScanStartMessage


class ScanStartMessageTestCase(TestCase):
    def test_constructor(self):
        message = ScanStartMessage(
            scan_id="scan_1",
            host_ip="1.1.1.1",
            host_name="foo",
            os_release="BarOS 1.0",
            package_list=["foo-1.2.3-1.x86_64"],
        )

        self.assertIsInstance(message.message_id, UUID)
        self.assertIsInstance(message.group_id, str)
        self.assertIsInstance(message.created, datetime)

        self.assertEqual(message.message_type, MessageType.SCAN_START)
        self.assertEqual(message.topic, "scanner/package/cmd/notus")

        self.assertEqual(message.scan_id, "scan_1")
        self.assertEqual(message.host_ip, "1.1.1.1")
        self.assertEqual(message.host_name, "foo")
        self.assertEqual(message.os_release, "BarOS 1.0")
        self.assertEqual(message.package_list, ["foo-1.2.3-1.x86_64"])

    def test_serialize(self):
        created = datetime.fromtimestamp(1628512774)
        message_id = UUID("63026767-029d-417e-9148-77f4da49f49a")
        group_id = "866350e8-1492-497e-b12b-c079287d51dd"
        message = ScanStartMessage(
            message_id=message_id,
            group_id=group_id,
            created=created,
            scan_id="scan_1",
            host_ip="1.1.1.1",
            host_name="foo",
            os_release="BarOS 1.0",
            package_list=["foo-1.2.3-1.x86_64"],
        )

        serialized = message.serialize()
        self.assertEqual(serialized["created"], 1628512774.0)
        self.assertEqual(
            serialized["message_id"], "63026767-029d-417e-9148-77f4da49f49a"
        )
        self.assertEqual(
            serialized["group_id"], "866350e8-1492-497e-b12b-c079287d51dd"
        )
        self.assertEqual(serialized["message_type"], "scan.start")
        self.assertEqual(serialized["scan_id"], "scan_1")
        self.assertEqual(serialized["host_ip"], "1.1.1.1")
        self.assertEqual(serialized["host_name"], "foo")
        self.assertEqual(serialized["os_release"], "BarOS 1.0")
        self.assertEqual(serialized["package_list"], ["foo-1.2.3-1.x86_64"])

    def test_deserialize(self):
        data = {
            "message_id": "63026767-029d-417e-9148-77f4da49f49a",
            "group_id": "866350e8-1492-497e-b12b-c079287d51dd",
            "created": 1628512774.0,
            "message_type": "scan.start",
            "scan_id": "scan_1",
            "host_ip": "1.1.1.1",
            "host_name": "foo",
            "os_release": "BarOS 1.0",
            "package_list": ["foo-1.2.3-1.x86_64"],
        }

        message = ScanStartMessage.deserialize(data)
        self.assertEqual(
            message.message_id, UUID("63026767-029d-417e-9148-77f4da49f49a")
        )
        self.assertEqual(
            message.group_id, "866350e8-1492-497e-b12b-c079287d51dd"
        )
        self.assertEqual(
            message.created,
            datetime.fromtimestamp(1628512774.0, tz=timezone.utc),
        )
        self.assertEqual(message.message_type, MessageType.SCAN_START)

        self.assertEqual(message.scan_id, "scan_1")
        self.assertEqual(message.host_ip, "1.1.1.1")
        self.assertEqual(message.host_name, "foo")
        self.assertEqual(message.os_release, "BarOS 1.0")
        self.assertEqual(message.package_list, ["foo-1.2.3-1.x86_64"])

        self.assertEqual(message.topic, "scanner/package/cmd/notus")

    def test_deserialize_invalid_message_type(self):
        data = {
            "message_id": "63026767-029d-417e-9148-77f4da49f49a",
            "group_id": "866350e8-1492-497e-b12b-c079287d51dd",
            "created": 1628512774.0,
            "message_type": "scan.status",
            "scan_id": "scan_1",
            "host_ip": "1.1.1.1",
            "host_name": "foo",
            "os_release": "BarOS 1.0",
            "package_list": ["foo-1.2.3-1.x86_64"],
        }

        with self.assertRaisesRegex(
            MessageParsingError,
            "Invalid message type MessageType.SCAN_STATUS for ScanStartMessage."
            " Must be MessageType.SCAN_START.",
        ):
            ScanStartMessage.deserialize(data)

    def test_deserialize_invalid_package_list(self):
        data = {
            "message_id": "63026767-029d-417e-9148-77f4da49f49a",
            "group_id": "866350e8-1492-497e-b12b-c079287d51dd",
            "created": "1628512774.0",
            "message_type": "scan.start",
            "scan_id": "scan_1",
            "host_ip": "1.1.1.1",
            "host_name": "foo",
            "os_release": "BarOS 1.0",
            "package_list": "foo-1.2.3-1.x86_64",
        }

        with self.assertRaisesRegex(
            MessageParsingError, "package_list must contain a list"
        ):
            ScanStartMessage.deserialize(data)



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest
from unittest.mock import patch

from notus.scanner.config import (
    DEFAULT_LOG_LEVEL,
    DEFAULT_MQTT_BROKER_ADDRESS,
    DEFAULT_MQTT_BROKER_PORT,
    DEFAULT_PID_FILE,
    DEFAULT_PRODUCTS_DIRECTORY,
    Config,
)
from notus.scanner.errors import ConfigFileError


class ConfigTestCase(unittest.TestCase):
    def test_config_defaults(self):
        config = Config()

        config_data = config.values()

        self.assertEqual(
            config_data.get("products-directory"),
            DEFAULT_PRODUCTS_DIRECTORY,
        )
        self.assertIsNone(config_data.get("log-file"))
        self.assertEqual(config_data.get("log-level"), DEFAULT_LOG_LEVEL)
        self.assertEqual(
            config_data.get("mqtt-broker-address"), DEFAULT_MQTT_BROKER_ADDRESS
        )
        self.assertEqual(
            config_data.get("mqtt-broker-port"), DEFAULT_MQTT_BROKER_PORT
        )
        self.assertEqual(config_data.get("pid-file"), DEFAULT_PID_FILE)

    @patch("pathlib.Path")
    def test_parsing_config_file(self, path_mock):
        path_mock.read_text.return_value = """[notus-scanner]
        products-directory = "/a/b"
        log-file = "/bar/foo"
        log-level = "DEBUG"
        mqtt-broker-address = "1.2.3.4"
        mqtt-broker-port = 1234
        pid-file = "/foo/bar"
        """

        config = Config()
        config.load(path_mock)

        config_data = config.values()

        self.assertEqual(config_data.get("products-directory"), "/a/b")
        self.assertEqual(config_data.get("log-file"), "/bar/foo")
        self.assertEqual(config_data.get("log-level"), "DEBUG")
        self.assertEqual(config_data.get("mqtt-broker-address"), "1.2.3.4")
        self.assertEqual(config_data.get("mqtt-broker-port"), 1234)
        self.assertEqual(config_data.get("pid-file"), "/foo/bar")

    @patch("pathlib.Path")
    def test_parsing_invalid_toml_file(self, path_mock):
        path_mock.read_text.return_value = """[notus-scanner]
        pid-file = /foo/bar
        """

        config = Config()

        with self.assertRaises(ConfigFileError):
            config.load(path_mock)

    @patch("pathlib.Path")
    def test_parsing_unknown_values(self, path_mock):
        path_mock.read_text.return_value = """[notus-scanner]
        products-directory = "/a/b"
        foo = "bar"
        """

        config = Config()
        config.load(path_mock)

        config_data = config.values()
        self.assertEqual(config_data.get("products-directory"), "/a/b")
        self.assertIsNone(config_data.get("foo"))

    @patch.dict(
        "os.environ",
        {
            "NOTUS_SCANNER_PRODUCTS_DIRECTORY": "/path/env/a",
            "NOTUS_SCANNER_LOG_FILE": "/path/env/b",
            "NOTUS_SCANNER_LOG_LEVEL": "WARN",
            "NOTUS_SCANNER_MQTT_BROKER_ADDRESS": "4.3.2.1",
            "NOTUS_SCANNER_MQTT_BROKER_PORT": "9876",
            "NOTUS_SCANNER_PID_FILE": "/path/env/c",
        },
        clear=True,
    )
    @patch("pathlib.Path")
    def test_parsing_environment_preference(self, path_mock):
        path_mock.read_text.return_value = """[notus-scanner]
        products-directory = "/path/config/a"
        log-file = "/path/config/b"
        log-level = "DEBUG"
        mqtt-broker-address = "1.2.3.4"
        mqtt-broker-port = 1234
        pid-file = "/path/config/c"
        """

        config = Config()
        config.load(path_mock)

        config_data = config.values()

        self.assertEqual(config_data.get("products-directory"), "/path/env/a")
        self.assertEqual(config_data.get("log-file"), "/path/env/b")
        self.assertEqual(config_data.get("log-level"), "WARN")
        self.assertEqual(config_data.get("mqtt-broker-address"), "4.3.2.1")
        self.assertEqual(config_data.get("mqtt-broker-port"), "9876")
        self.assertEqual(config_data.get("pid-file"), "/path/env/c")



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from .packages.package import Package, PackageAdvisory


class Vulnerability:
    """Represents a vulnerable package on a host"""

    vulnerability: Dict[Package, List[PackageAdvisory]]

    def __init__(self) -> None:
        self.vulnerability = {}

    def add(self, package: Package, fixed: PackageAdvisory) -> None:
        """Adds a detection to a package. A detection indicates a vulnerable
        package. Multiple detections indicates a version range."""
        if package not in self.vulnerability:
            self.vulnerability[package] = list()
        self.vulnerability[package].append(fixed)

    def combine(self, vul: "Vulnerability"):
        self.vulnerability = {**self.vulnerability, **vul.vulnerability}

    def get(self) -> Dict[Package, List[PackageAdvisory]]:
        return self.vulnerability


class Vulnerabilities:
    """Adds a vulnerable package corresponding to a given OID. It is
    possible, that there are multiple vulnerable packages for a single
    OID"""

    vulnerabilities: Dict[str, Vulnerability]

    def __init__(self) -> None:
        self.vulnerabilities = dict()

    def __len__(self) -> int:
        return len(self.vulnerabilities)

    def add(self, oid: str, vul: Vulnerability):
        if not oid in self.vulnerabilities:
            self.vulnerabilities[oid] = vul
            return

        self.vulnerabilities[oid].combine(vul)

    def get(self) -> Dict[str, Vulnerability]:
        return self.vulnerabilities



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from .deb import DEBPackage
from .ebuild import EBuildPackage
from .package import Package, PackageType
from .rpm import RPMPackage
from .slackware import SlackPackage

__SWITCHER = {
    PackageType.RPM: RPMPackage,
    PackageType.DEB: DEBPackage,
    PackageType.EBUILD: EBuildPackage,
    PackageType.SLACK: SlackPackage,
}


def package_class_by_type(pt: PackageType) -> Optional[Package]:
    """
    package_class_by_type may returns Package if defined otherwise None
    """
    return __SWITCHER.get(pt)



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
import logging
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Callable, Dict, Optional

from ..errors import AdvisoriesLoadingError
from ..models.packages import package_class_by_type
from ..models.packages.package import PackageAdvisories, PackageType
from .gpg_sha_verifier import VerificationResult
from .loader import AdvisoriesLoader

logger = logging.getLogger(__name__)


def _get_operating_system_file_name(operating_system: str) -> str:
    return operating_system.strip().replace(" ", "_").lower()


class JSONAdvisoriesLoader(AdvisoriesLoader):
    def __init__(
        self,
        advisories_directory_path: Path,
        verify: Callable[[Path], VerificationResult],
    ):
        self._advisories_directory_path = advisories_directory_path
        self._verify = verify

    def __load_data(self, operating_system: str) -> Optional[Dict]:
        os_file_name = _get_operating_system_file_name(operating_system)
        json_file_path = (
            self._advisories_directory_path / f"{os_file_name}.notus"
        )
        # since the data is coming from the outside it should not crash
        # on wrongfully send data instead print a warning and return None
        if not json_file_path.exists():
            logger.log(
                logging.WARNING,
                "Could not load advisories from %s. File does not exist.",
                json_file_path.absolute(),
            )
            return None
        # If there is a file but unable to verify it could be that the feed
        # is corrupted and the application should stop
        verify_result = self._verify(json_file_path)
        if verify_result == VerificationResult.SUCCESS:
            logger.debug("File '%s' verification successful.", json_file_path)
        else:
            reason = (
                "File verification failed."
                if verify_result != VerificationResult.INVALID_NAME
                else "OS name does not match filename."
            )
            raise AdvisoriesLoadingError(
                f"could not load advisories from {json_file_path.absolute()}. "
                f"{reason}"
            )

        if json_file_path.stat().st_size < 2:
            # the minimum size of a json file is 2 bytes ({} or [])
            return None

        with json_file_path.open("r", encoding="utf-8") as f:
            try:
                return json.load(f)
            except JSONDecodeError as e:
                raise AdvisoriesLoadingError(
                    "Could not load advisories from "
                    f"{json_file_path.absolute()}. Error in line {e.lineno} "
                    "while decoding JSON data."
                ) from None

    def load_package_advisories(
        self, operating_system: str
    ) -> Optional[PackageAdvisories]:
        data = self.__load_data(operating_system=operating_system)
        if not data:
            return None
        package_type_id = data.get("package_type", "")
        try:
            package_type = PackageType(package_type_id)
        except ValueError:
            logger.warning("%s invalid package type.", package_type_id)
            return None
        package_advisories = PackageAdvisories(package_type)
        package_class = package_class_by_type(package_type)
        if not package_class:
            logger.warning("%s has no package implementation", package_type_id)
            return None

        for advisory_data in data.get("advisories", []):
            if not "oid" in advisory_data:
                logger.error("No OID found for JSON advisory %s", advisory_data)
                continue

            try:
                # parse required data
                oid = advisory_data["oid"]
                fixed_packages = advisory_data["fixed_packages"]
            except (KeyError, TypeError) as e:
                logger.warning(
                    "Error while parsing %s for %s. Error was %s",
                    advisory_data,
                    operating_system,
                    e,
                )
                continue

            advisory = oid

            for package_dict in fixed_packages:
                full_name = package_dict.get("full_name")
                if full_name:
                    package = package_class.from_full_name(full_name)
                else:
                    package = package_class.from_name_and_full_version(
                        package_dict.get("name"),
                        package_dict.get("full_version"),
                    )
                if not package:
                    logger.warning(
                        "Could not parse fixed package information from %s "
                        "in %s",
                        package_dict,
                        operating_system,
                    )
                    continue

                package_advisories.add_advisory_for_package(
                    package, advisory, package_dict.get("specifier")
                )

        return package_advisories



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from enum import Enum
from typing import Any, Dict, Optional, Union
from uuid import UUID

from ..errors import MessageParsingError
from .message import Message, MessageType


class ResultType(Enum):
    ALARM = "ALARM"


class ResultMessage(Message):
    message_type: MessageType = MessageType.RESULT
    topic = "scanner/scan/info"

    def __init__(
        self,
        *,
        scan_id: str,
        host_ip: str,
        host_name: str,
        oid: str,
        value: str,
        port: str = "package",
        uri: str = None,
        result_type: ResultType = ResultType.ALARM,
        message_id: Optional[UUID] = None,
        group_id: Optional[UUID] = None,
        created: Optional[datetime] = None,
    ):
        super().__init__(
            message_id=message_id, group_id=group_id, created=created
        )
        self.scan_id = scan_id
        self.host_ip = host_ip
        self.host_name = host_name
        self.oid = oid
        self.value = value
        self.port = port
        self.uri = uri
        self.result_type = result_type

    def serialize(self) -> Dict[str, Union[int, str]]:
        message = super().serialize()
        message.update(
            {
                "scan_id": self.scan_id,
                "host_ip": self.host_ip,
                "host_name": self.host_name,
                "oid": self.oid,
                "value": self.value,
                "port": self.port,
                "uri": self.uri,
                "result_type": self.result_type.value,
            }
        )
        return message

    @classmethod
    def _parse(cls, data: Dict[str, Union[int, str]]) -> Dict[str, Any]:
        kwargs = super()._parse(data)
        try:
            kwargs.update(
                {
                    "scan_id": data.get("scan_id"),
                    "host_ip": data.get("host_ip"),
                    "host_name": data.get("host_name"),
                    "oid": data.get("oid"),
                    "value": data.get("value"),
                    "port": data.get("port"),
                    "uri": data.get("uri"),
                    "result_type": ResultType(data.get("result_type")),
                }
            )
        except ValueError as e:
            raise MessageParsingError(
                f"error while parsing 'result_type', {e}"
            ) from e
        return kwargs



# Copyright (C) 2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
import unittest
from pathlib import Path
from typing import List, Optional, Tuple

from notus.scanner.loader.gpg_sha_verifier import VerificationResult
from notus.scanner.loader.json import JSONAdvisoriesLoader
from notus.scanner.messages.message import Message
from notus.scanner.messages.start import ScanStartMessage
from notus.scanner.messaging.publisher import Publisher
from notus.scanner.scanner import NotusScanner

_here = Path(__file__).parent


class FakePublisher(Publisher):
    def __init__(self):
        self.results = []

    def publish(self, message: Message) -> None:
        serialized = message.serialize()
        values = str(serialized.get("value", "")).split("\n")
        for value in values:
            if value.find("Installed version:") >= 0:
                installed = value.split(":")[1].strip()
                self.results.append(installed)


class VerifierTestCase(unittest.TestCase):
    """
    VerifierTestCase loads a notus advisory for FakeSpecifierOS
    and iterates through defined fixed packages and their specifier
    to generate and test them.
    """

    def per_symbol(
        self, name: str, verifier: Optional[str]
    ) -> Tuple[List[str], List[str]]:
        """
        returns not_in_result and in result
        """
        greater = lambda: name.replace("15", "16")
        smaller = lambda: name.replace("15", "14")
        if not verifier:
            return [greater(), name], [smaller()]
        if verifier == ">":
            return [greater()], [name, smaller()]
        if verifier == ">=":
            return [greater(), name], [smaller()]
        if verifier == "<":
            return [smaller()], [name, greater()]
        if verifier == "<=":
            return [smaller(), name], [greater()]
        return [name], [smaller(), greater()]

    def generate_test_cases(self):
        jdict = json.loads((_here / "fakespecifier_os.notus").read_bytes())
        fixed_packages = (
            f.get("fixed_packages", []) for f in jdict.get("advisories", {})
        )
        cases = (
            (fp.get("full_name"), fp.get("specifier"))
            for fps in fixed_packages
            for fp in fps
        )
        # cases that should not appear in result
        not_in = []
        # cases that should appear in result
        is_in = []
        for case in cases:
            c_not_in, c_is_in = self.per_symbol(*case)
            not_in = not_in + c_not_in
            is_in = is_in + c_is_in
        # packagelist is both combined
        return not_in + is_in, not_in, is_in

    def test_verifier(self):
        loader = JSONAdvisoriesLoader(
            advisories_directory_path=_here,
            verify=lambda _: VerificationResult.SUCCESS,
        )
        publisher = FakePublisher()
        scanner = NotusScanner(loader, publisher)
        pkg_list, not_in_results, in_results = self.generate_test_cases()
        msg = ScanStartMessage(
            scan_id="scanus praktikus",
            os_release="FakeSpecifier OS",
            host_ip="127.0.0.2",
            host_name="localhorst",
            package_list=pkg_list,
        )
        scanner.run_scan(msg)
        results = set(publisher.results)

        self.assertEqual(len(results), len(publisher.results))

        self.assertEqual(
            set(),
            results.intersection(not_in_results),
        )

        self.assertEqual(set(in_results), results.intersection(in_results))



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase

from notus.scanner.models.packages.slackware import SlackPackage


class SlackPackageTestCase(TestCase):
    def test_compare_gt(self):
        """packages should be comparable"""
        package1 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="4",
            arch="x86_64",
            target="15.0",
            full_name="foo-bar-1.2.3-x86_64-4_slack15.0",
            full_version="1.2.3-x86_64-4_slack15.0",
        )
        package2 = SlackPackage(
            name="foo-bar",
            version="1.2.4",
            build="4",
            arch="x86_64",
            target="15.0",
            full_name="foo-bar-1.2.4-x86_64-4_slack15.0",
            full_version="1.2.4-x86_64-4_slack15.0",
        )
        self.assertGreater(package2, package1)

        package2 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="5",
            arch="x86_64",
            target="15.0",
            full_name="foo-bar-1.2.3-x86_64-5_slack15.0",
            full_version="1.2.3-x86_64-5_slack15.0",
        )
        self.assertGreater(package2, package1)

        package2 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="4",
            arch="x86_64",
            target="15.1",
            full_name="foo-bar-1.2.3-x86_64-4_slack15.1",
            full_version="1.2.3-x86_64-4_slack15.1",
        )
        self.assertGreater(package2, package1)

    def test_compare_gt_different_architecture(self):
        """packages of different architecture should not be comparable"""
        package1 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="4",
            arch="x86_64",
            target="15.0",
            full_name="foo-bar-1.2.3-x86_64-4_slack15.0",
            full_version="1.2.3-x86_64-4_slack15.0",
        )
        package2 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="4",
            arch="aarch64",
            target="15.0",
            full_name="foo-bar-1.2.3-aarch64-4_slack15.0",
            full_version="1.2.3-aarch64-4_slack15.0",
        )
        package3 = SlackPackage(
            name="foo-bar",
            version="1.2.4",
            build="4",
            arch="aarch64",
            target="15.0",
            full_name="foo-bar-1.2.4-aarch64-4_slack15.0",
            full_version="1.2.4-aarch64-4_slack15.0",
        )
        package4 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="5",
            arch="aarch64",
            target="15.0",
            full_name="foo-bar-1.2.3-aarch64-5_slack15.0",
            full_version="1.2.3-aarch64-5_slack15.0",
        )
        package5 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="4",
            arch="aarch64",
            target="15.1",
            full_name="foo-bar-1.2.3-aarch64-4_slack15.1",
            full_version="1.2.3-aarch64-4_slack15.1",
        )
        self.assertFalse(package2 > package1)
        self.assertFalse(package1 > package2)
        self.assertFalse(package3 > package1)
        self.assertFalse(package1 > package3)
        self.assertFalse(package4 > package1)
        self.assertFalse(package1 > package4)
        self.assertFalse(package5 > package1)
        self.assertFalse(package1 > package5)

    def test_compare_gt_different_name(self):
        """different packagtes should not be comparable"""
        package1 = SlackPackage(
            name="foo",
            version="1.2.3",
            build="4",
            arch="x86_64",
            target="15.0",
            full_name="foo-1.2.3-x86_64-4_slack15.0",
            full_version="1.2.3-x86_64-4_slack15.0",
        )
        package2 = SlackPackage(
            name="bar",
            version="1.2.3",
            build="4",
            arch="x86_64",
            target="15.0",
            full_name="bar-1.2.3-x86_64-4_slack15.0",
            full_version="1.2.3-x86_64-4_slack15.0",
        )
        self.assertFalse(package2 > package1)
        self.assertFalse(package1 > package2)

    def test_compare_less(self):
        """packages should be comparable"""
        package1 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="4",
            arch="x86_64",
            target="15.0",
            full_name="foo-bar-1.2.3-x86_64-4_slack15.0",
            full_version="1.2.3-x86_64-4_slack15.0",
        )
        package2 = SlackPackage(
            name="foo-bar",
            version="1.2.4",
            build="4",
            arch="x86_64",
            target="15.0",
            full_name="foo-bar-1.2.4-x86_64-4_slack15.0",
            full_version="1.2.4-x86_64-4_slack15.0",
        )
        self.assertLess(package1, package2)

        package2 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="5",
            arch="x86_64",
            target="15.0",
            full_name="foo-bar-1.2.3-x86_64-5_slack15.0",
            full_version="1.2.3-x86_64-5_slack15.0",
        )
        self.assertLess(package1, package2)

        package2 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="4",
            arch="x86_64",
            target="15.1",
            full_name="foo-bar-1.2.3-x86_64-4_slack15.1",
            full_version="1.2.3-x86_64-4_slack15.1",
        )
        self.assertLess(package1, package2)

    def test_compare_equal(self):
        """packages with the same data should be equal"""
        package1 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="4",
            arch="x86_64",
            target="15.0",
            full_name="foo-bar-1.2.3-x86_64-4_slack15.0",
            full_version="1.2.3-x86_64-4_slack15.0",
        )
        package2 = SlackPackage(
            name="foo-bar",
            version="1.2.3",
            build="4",
            arch="x86_64",
            target="15.0",
            full_name="foo-bar-1.2.3-x86_64-4_slack15.0",
            full_version="1.2.3-x86_64-4_slack15.0",
        )

        self.assertEqual(package1, package2)

    def test_from_full_name(self):
        """it should be possible to create packages via the full name"""
        self.assertIsNone(SlackPackage.from_full_name(None))

        package = SlackPackage.from_full_name("flac-1.3.4-foo-1_slack15.0")
        self.assertEqual(package.arch, "foo")

        package = SlackPackage.from_full_name("flac-1.3.4-x86_64-1_slack15.0")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "flac")
        self.assertEqual(package.version, "1.3.4")
        self.assertEqual(package.build, "1")
        self.assertEqual(package.target, "15.0")
        self.assertEqual(package.full_version, "1.3.4-x86_64-1_slack15.0")
        self.assertEqual(package.full_name, "flac-1.3.4-x86_64-1_slack15.0")

        package = SlackPackage.from_full_name("kernel-source-5.15.27-noarch-1")
        self.assertEqual(package.arch, "noarch")
        self.assertEqual(package.name, "kernel-source")
        self.assertEqual(package.version, "5.15.27")
        self.assertEqual(package.build, "1")
        self.assertEqual(package.target, "")
        self.assertEqual(package.full_version, "5.15.27-noarch-1")
        self.assertEqual(package.full_name, "kernel-source-5.15.27-noarch-1")

        package = SlackPackage.from_full_name("libjpeg-v8a-x86_64-2")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "libjpeg")
        self.assertEqual(package.version, "v8a")
        self.assertEqual(package.build, "2")
        self.assertEqual(package.target, "")
        self.assertEqual(package.full_version, "v8a-x86_64-2")
        self.assertEqual(package.full_name, "libjpeg-v8a-x86_64-2")

        package = SlackPackage.from_full_name(" libjpeg-v8a-x86_64-2\r\n")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "libjpeg")
        self.assertEqual(package.version, "v8a")
        self.assertEqual(package.build, "2")
        self.assertEqual(package.target, "")
        self.assertEqual(package.full_version, "v8a-x86_64-2")
        self.assertEqual(package.full_name, "libjpeg-v8a-x86_64-2")

        package = SlackPackage.from_full_name("libjpeg-v8a-x86_64")
        self.assertIsNone(package)

        package = SlackPackage.from_full_name("libjpeg-v8a-foo-2")
        self.assertEqual(package.arch, "foo")

    def test_from_name_and_full_version(self):
        """it should be possible to create packages from name and full
        version"""
        self.assertIsNone(SlackPackage.from_name_and_full_version(None, None))

        package = SlackPackage.from_name_and_full_version(
            "flac", "1.3.4-x86_64-1_slack15.0"
        )
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "flac")
        self.assertEqual(package.version, "1.3.4")
        self.assertEqual(package.build, "1")
        self.assertEqual(package.target, "15.0")
        self.assertEqual(package.full_version, "1.3.4-x86_64-1_slack15.0")
        self.assertEqual(package.full_name, "flac-1.3.4-x86_64-1_slack15.0")

        package = SlackPackage.from_name_and_full_version(
            "flac", "1.3.4-x86_64"
        )
        self.assertIsNone(package)

        package = SlackPackage.from_name_and_full_version("flac", "1.3.4-foo-1")
        self.assertEqual(package.arch, "foo")



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase

from notus.scanner.models.packages.rpm import RPMPackage


class RPMPackageTestCase(TestCase):
    def test_compare_gt(self):
        """packages should be comparable"""
        package1 = RPMPackage(
            name="foo-bar",
            version="1.2.3",
            release="4",
            arch="x86_64",
            full_name="foo-bar-1.2.3-4.x86_64",
            full_version="1.2.3-4.x86_64",
        )
        package2 = RPMPackage(
            name="foo-bar",
            version="1.2.4",
            release="4",
            arch="x86_64",
            full_name="foo-bar-1.2.4-4.x86_64",
            full_version="1.2.4-4.x86_64",
        )
        self.assertGreater(package2, package1)

        package2 = RPMPackage(
            name="foo-bar",
            version="1.2.3",
            release="5",
            arch="x86_64",
            full_name="foo-bar-1.2.3-5.x86_64",
            full_version="1.2.3-5.x86_64",
        )
        self.assertGreater(package2, package1)

    def test_compare_gt_different_architecture(self):
        """packages of different architecture should not be comparable"""
        package1 = RPMPackage(
            name="foo-bar",
            version="1.2.3",
            release="4",
            arch="x86_64",
            full_name="foo-bar-1.2.3-4.x86_64",
            full_version="1.2.3-4.x86_64",
        )
        package2 = RPMPackage(
            name="foo-bar",
            version="1.2.3",
            release="4",
            arch="aarch64",
            full_name="foo-bar-1.2.3-4.aarch64",
            full_version="1.2.3-4.aarch64",
        )
        package3 = RPMPackage(
            name="foo-bar",
            version="1.2.4",
            release="4",
            arch="aarch64",
            full_name="foo-bar-1.2.4-4.aarch64",
            full_version="1.2.4-4.aarch64",
        )
        package4 = RPMPackage(
            name="foo-bar",
            version="1.2.3",
            release="5",
            arch="aarch64",
            full_name="foo-bar-1.2.3-5.aarch64",
            full_version="1.2.3-5.aarch64",
        )
        self.assertFalse(package2 > package1)
        self.assertFalse(package1 > package2)
        self.assertFalse(package3 > package1)
        self.assertFalse(package1 > package3)
        self.assertFalse(package4 > package1)
        self.assertFalse(package1 > package4)

    def test_compare_gt_different_name(self):
        """different packagtes should not be comparable"""
        package1 = RPMPackage(
            name="foo",
            version="1.2.3",
            release="4",
            arch="x86_64",
            full_name="foo-1.2.3-4.x86_64",
            full_version="1.2.3-4.x86_64",
        )
        package2 = RPMPackage(
            name="bar",
            version="1.2.3",
            release="4",
            arch="x86_64",
            full_name="bar-1.2.3-4.x86_64",
            full_version="1.2.3-4.x86_64",
        )
        self.assertFalse(package2 > package1)
        self.assertFalse(package1 > package2)

    def test_compare_less(self):
        """packages should be comparable"""
        package1 = RPMPackage(
            name="foo-bar",
            version="1.2.3",
            release="4",
            arch="x86_64",
            full_name="foo-bar-1.2.3-4.x86_64",
            full_version="1.2.3-4.x86_64",
        )
        package2 = RPMPackage(
            name="foo-bar",
            version="1.2.4",
            release="4",
            arch="x86_64",
            full_name="foo-bar-1.2.4-4.x86_64",
            full_version="1.2.4-4.x86_64",
        )
        self.assertLess(package1, package2)

        package2 = RPMPackage(
            name="foo-bar",
            version="1.2.3",
            release="5",
            arch="x86_64",
            full_name="foo-bar-1.2.3-5.x86_64",
            full_version="1.2.3-5.x86_64",
        )
        self.assertLess(package1, package2)

    def test_compare_equal(self):
        """packages with the same data should be equal"""
        package1 = RPMPackage(
            name="foo-bar",
            version="1.2.3",
            release="4",
            arch="x86_64",
            full_name="foo-bar-1.2.3-4.x86_64",
            full_version="1.2.3-4.x86_64",
        )
        package2 = RPMPackage(
            name="foo-bar",
            version="1.2.3",
            release="4",
            arch="x86_64",
            full_name="foo-bar-1.2.3-4.x86_64",
            full_version="1.2.3-4.x86_64",
        )

        self.assertEqual(package1, package2)

    def test_from_full_name(self):
        """it should be possible to create packages via the full name"""
        self.assertIsNone(RPMPackage.from_full_name(None))

        package = RPMPackage.from_full_name("keyutils-1.5.8-3.foo")
        self.assertEqual(package.arch, "foo")

        package = RPMPackage.from_full_name("keyutils-1.5.8-3")
        self.assertEqual(package.arch, "")

        package = RPMPackage.from_full_name(
            "mesa-libgbm-11.2.2-2.20160614.x86_64"
        )

        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "mesa-libgbm")
        self.assertEqual(package.version, "11.2.2")
        self.assertEqual(package.release, "2.20160614")
        self.assertEqual(
            package.full_name, "mesa-libgbm-11.2.2-2.20160614.x86_64"
        )

        package = RPMPackage.from_full_name("keyutils-1.5.8-3.x86_64")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "keyutils")
        self.assertEqual(package.version, "1.5.8")
        self.assertEqual(package.release, "3")
        self.assertEqual(package.full_name, "keyutils-1.5.8-3.x86_64")

        package = RPMPackage.from_full_name(
            "httpd-manual-2.4.6-45.0.1.4.h10.noarch"
        )
        self.assertEqual(package.arch, "noarch")
        self.assertEqual(package.name, "httpd-manual")
        self.assertEqual(package.version, "2.4.6")
        self.assertEqual(package.release, "45.0.1.4.h10")
        self.assertEqual(
            package.full_name, "httpd-manual-2.4.6-45.0.1.4.h10.noarch"
        )

        package = RPMPackage.from_full_name("cups-libs-1.6.3-26.h1.x86_64")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "cups-libs")
        self.assertEqual(package.version, "1.6.3")
        self.assertEqual(package.release, "26.h1")
        self.assertEqual(package.full_name, "cups-libs-1.6.3-26.h1.x86_64")

        package = RPMPackage.from_full_name("GConf2-3.2.6-8.x86_64")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "GConf2")
        self.assertEqual(package.version, "3.2.6")
        self.assertEqual(package.release, "8")
        self.assertEqual(package.full_name, "GConf2-3.2.6-8.x86_64")

        package = RPMPackage.from_full_name("libtool-ltdl-2.4.2-21.x86_64")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "libtool-ltdl")
        self.assertEqual(package.version, "2.4.2")
        self.assertEqual(package.release, "21")
        self.assertEqual(package.full_name, "libtool-ltdl-2.4.2-21.x86_64")

        package = RPMPackage.from_full_name("microcode_ctl-2.1-22.6.h2.x86_64")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "microcode_ctl")
        self.assertEqual(package.version, "2.1")
        self.assertEqual(package.release, "22.6.h2")
        self.assertEqual(package.full_name, "microcode_ctl-2.1-22.6.h2.x86_64")

        package = RPMPackage.from_full_name("postgresql-libs-9.2.23-3.x86_64")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "postgresql-libs")
        self.assertEqual(package.version, "9.2.23")
        self.assertEqual(package.release, "3")
        self.assertEqual(package.full_name, "postgresql-libs-9.2.23-3.x86_64")

        package = RPMPackage.from_full_name("NetworkManager-1.8.0-9.h2.x86_64")
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "NetworkManager")
        self.assertEqual(package.version, "1.8.0")
        self.assertEqual(package.release, "9.h2")
        self.assertEqual(package.full_name, "NetworkManager-1.8.0-9.h2.x86_64")

        package = RPMPackage.from_full_name(
            "perl-Pod-Escapes-1.04-285.h2.noarch"
        )
        self.assertEqual(package.arch, "noarch")
        self.assertEqual(package.name, "perl-Pod-Escapes")
        self.assertEqual(package.version, "1.04")
        self.assertEqual(package.release, "285.h2")
        self.assertEqual(
            package.full_name, "perl-Pod-Escapes-1.04-285.h2.noarch"
        )

        package = RPMPackage.from_full_name(" libtool-ltdl-2.4.2-21.x86_64\r\n")
        self.assertEqual(package.arch, "x86_64")

    def test_from_name_and_full_version(self):
        """it should be possible to create packages from name and full
        version"""
        self.assertIsNone(RPMPackage.from_name_and_full_version(None, None))

        package = RPMPackage.from_name_and_full_version(
            "cups-libs", "1.6.3-26.h1.x86_64"
        )
        self.assertEqual(package.arch, "x86_64")
        self.assertEqual(package.name, "cups-libs")
        self.assertEqual(package.version, "1.6.3")
        self.assertEqual(package.release, "26.h1")
        self.assertEqual(package.full_name, "cups-libs-1.6.3-26.h1.x86_64")

    def test_exceptions(self):
        """tests for the exceptions _fips and .ksplice"""
        package1 = RPMPackage.from_full_name("gnutls-3.6.16-4.el8.x86_64")
        package2 = RPMPackage.from_full_name(
            "gnutls-3.6.16-4.0.1.el8_fips.x86_64"
        )

        self.assertFalse(package1 > package2)
        self.assertFalse(package2 > package1)

        package1 = RPMPackage.from_full_name("gnutls-3.6.16-4.el8_fips.x86_64")

        self.assertTrue(package2 > package1)

        package1 = RPMPackage.from_full_name(
            "openssl-libs-1.0.2k-24.0.3.el7_8.x86_64"
        )
        package2 = RPMPackage.from_full_name(
            "openssl-libs-1.0.2k-24.0.3.ksplice1.el7_9.x86_64"
        )

        self.assertFalse(package1 > package2)
        self.assertFalse(package2 > package1)

        package1 = RPMPackage.from_full_name(
            "openssl-libs-1.0.2k-24.0.3.ksplice1.el7_8.x86_64"
        )

        self.assertTrue(package2 > package1)



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from unittest import TestCase, mock
from uuid import UUID

from notus.scanner.messages.start import ScanStartMessage
from notus.scanner.messaging.mqtt import (
    MQTTDaemon,
    MQTTPublisher,
    MQTTSubscriber,
)


class MQTTPublisherTestCase(TestCase):
    def test_publish(self):
        client = mock.MagicMock()
        publisher = MQTTPublisher(client)

        created = datetime.fromtimestamp(1628512774)
        message_id = UUID("63026767-029d-417e-9148-77f4da49f49a")
        group_id = UUID("866350e8-1492-497e-b12b-c079287d51dd")
        message = ScanStartMessage(
            message_id=message_id,
            group_id=group_id,
            created=created,
            scan_id="scan_1",
            host_ip="1.1.1.1",
            host_name="foo",
            os_release="BarOS 1.0",
            package_list=["foo-1.2.3-1.x86_64"],
        )

        publisher.publish(message)

        client.publish.assert_called_with(
            "scanner/package/cmd/notus",
            '{"message_id": "63026767-029d-417e-9148-77f4da49f49a", '
            '"message_type": "scan.start", '
            '"group_id": "866350e8-1492-497e-b12b-c079287d51dd", '
            '"created": 1628512774.0, '
            '"scan_id": "scan_1", '
            '"host_ip": "1.1.1.1", '
            '"host_name": "foo", '
            '"os_release": "BarOS 1.0", '
            '"package_list": ["foo-1.2.3-1.x86_64"]}',
            qos=1,
        )


class MQTTSubscriberTestCase(TestCase):
    def test_subscribe(self):
        client = mock.MagicMock()
        callback = mock.MagicMock()
        callback.__name__ = "callback_name"

        subscriber = MQTTSubscriber(client)

        subscriber.subscribe(ScanStartMessage, callback)

        client.subscribe.assert_called_with("scanner/package/cmd/notus", qos=1)


class MQTTDaemonTestCase(TestCase):
    def test_connect(self):
        client = mock.MagicMock()

        # pylint: disable=unused-variable
        daemon = MQTTDaemon(client)

        client.connect.assert_called_with()

    def test_run(self):
        client = mock.MagicMock()

        daemon = MQTTDaemon(client)

        daemon.run()

        client.loop_forever.assert_called_with()



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase

from notus.scanner.models.packages.package import PackageAdvisory
from notus.scanner.models.packages.rpm import RPMPackage
from notus.scanner.models.vulnerability import Vulnerabilities, Vulnerability


class VulnerabilityTestCase(TestCase):
    def test_vulnerability(self):
        vul = Vulnerability()

        package = RPMPackage.from_full_name(
            "foo-1.2.3-4.x86_64",
        )
        if not package:
            self.fail("Cannot build package")
        fixed_package = RPMPackage.from_full_name(
            "foo-1.2.4-4.x86_64",
        )
        if not fixed_package:
            self.fail("Cannot build package")

        fixed = PackageAdvisory(
            fixed_package, "1.2.3.4.5", ">=", lambda _: False
        )

        vul.add(package, fixed)

        for key, val in vul.get().items():
            self.assertEqual(package, key)
            self.assertListEqual([fixed], val)

        fixed_package = RPMPackage.from_full_name(
            "foo-1.2.2-4.x86_64",
        )
        if not fixed_package:
            self.fail("Cannot build package")

        fixed = PackageAdvisory(
            fixed_package, "1.2.3.4.5", "<", lambda _: False
        )

        vul.add(package, fixed)

        self.assertEqual(len(vul.get()), 1)
        self.assertEqual(len(vul.get()[package]), 2)

        vul2 = Vulnerability()

        package = RPMPackage.from_full_name(
            "bar-1.2.3-4.x86_64",
        )
        if not package:
            self.fail("Cannot build package")

        fixed_package = RPMPackage.from_full_name(
            "bar-1.2.4-4.x86_64",
        )
        if not fixed_package:
            self.fail("Cannot build package")

        fixed = PackageAdvisory(
            fixed_package, "1.2.3.4.5", ">=", lambda _: False
        )

        vul2.add(package, fixed)
        vul.combine(vul2)

        self.assertEqual(len(vul.get()), 2)


class VulnerabilitiesTestCase(TestCase):
    def test_vulnerabilities(self):
        vuls = Vulnerabilities()
        vul = Vulnerability()

        package = RPMPackage.from_full_name(
            "foo-1.2.3-4.x86_64",
        )
        if not package:
            self.fail("Cannot build package")

        fixed_package = RPMPackage.from_full_name(
            "foo-1.2.4-4.x86_64",
        )
        if not fixed_package:
            self.fail("Cannot build package")

        fixed = PackageAdvisory(
            fixed_package, "1.2.3.4.5", ">=", lambda _: False
        )

        vul.add(package, fixed)
        vuls.add("1.2.3.4.5", vul)

        self.assertEqual(len(vuls), 1)
        self.assertTrue("1.2.3.4.5" in vuls.get())
        self.assertEqual(vuls.get()["1.2.3.4.5"], vul)

        vul = Vulnerability()

        package = RPMPackage.from_full_name(
            "bar-1.2.3-4.x86_64",
        )
        if not package:
            self.fail("Cannot build package")

        fixed_package = RPMPackage.from_full_name(
            "bar-1.2.4-4.x86_64",
        )
        if not fixed_package:
            self.fail("Cannot build package")

        fixed = PackageAdvisory(
            fixed_package, "1.2.3.4.6", ">=", lambda _: False
        )

        vul.add(package, fixed)
        vuls.add("1.2.3.4.6", vul)

        self.assertEqual(len(vuls), 2)
        self.assertTrue("1.2.3.4.6" in vuls.get())
        self.assertEqual(vuls.get()["1.2.3.4.6"], vul)



# Copyright (C) 2014-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys
from pathlib import Path
from typing import Callable, Dict, Optional

from .__version__ import __version__
from .cli import CliParser
from .errors import AdvisoriesLoadingError, Sha256SumLoadingError
from .loader import JSONAdvisoriesLoader
from .loader.gpg_sha_verifier import (
    ReloadConfiguration,
    VerificationResult,
    create_verify,
    reload_sha256sums,
)
from .messages.start import ScanStartMessage
from .messaging.mqtt import (
    MQTTClient,
    MQTTDaemon,
    MQTTPublisher,
    MQTTSubscriber,
)
from .scanner import NotusScanner
from .utils import (
    create_pid,
    go_to_background,
    init_logging,
    init_signal_handler,
)

logger = logging.getLogger(__name__)

SENTRY_DSN_NOTUS_SCANNER = os.environ.get("SENTRY_DSN_NOTUS_SCANNER")
if SENTRY_DSN_NOTUS_SCANNER:
    import sentry_sdk  # pylint: disable=import-error

    sentry_sdk.init(  # pylint: disable=abstract-class-instantiated
        SENTRY_DSN_NOTUS_SCANNER,
        traces_sample_rate=1.0,
        server_name=os.environ.get("SENTRY_SERVER_NAME"),
        environment=os.environ.get("SENTRY_ENVIRONMENT"),
    )


def hashsum_verificator(
    products_directory_path: Path, disable: bool
) -> Callable[[Path], VerificationResult]:
    if disable:
        logger.info("hashsum verification is disabled")
        return lambda _: VerificationResult.SUCCESS

    def on_hash_sum_verification_failure(
        _: Optional[Dict[str, str]]
    ) -> Dict[str, str]:
        raise Sha256SumLoadingError(
            f"Unable to verify signature of {sha_sum_file_path}"
        )

    sha_sum_file_path = products_directory_path / "sha256sums"
    sha_sum_reload_config = ReloadConfiguration(
        hash_file=sha_sum_file_path,
        on_verification_failure=on_hash_sum_verification_failure,
    )

    sums = reload_sha256sums(sha_sum_reload_config)
    return create_verify(sums)


def run_daemon(
    mqtt_broker_address: str,
    mqtt_broker_port: int,
    products_directory_path: Path,
    disable_hashsum_verification: bool,
):
    """Initialize the mqtt client, mqtt handler, notus scanner and run
    forever
    """

    if not products_directory_path.is_dir():
        raise AdvisoriesLoadingError(
            f"Can't load advisories. {products_directory_path.absolute()} is"
            " not a directory."
        )

    verifier = hashsum_verificator(
        products_directory_path, disable_hashsum_verification
    )

    loader = JSONAdvisoriesLoader(
        advisories_directory_path=products_directory_path, verify=verifier
    )
    client = MQTTClient(
        mqtt_broker_address=mqtt_broker_address,
        mqtt_broker_port=mqtt_broker_port,
    )
    daemon: MQTTDaemon
    try:
        daemon = MQTTDaemon(client)
    except ConnectionRefusedError:
        logger.error(
            "Could not connect to MQTT broker at %s. Connection refused.",
            mqtt_broker_address,
        )
        sys.exit(1)

    publisher = MQTTPublisher(client)
    scanner = NotusScanner(loader=loader, publisher=publisher)

    subscriber = MQTTSubscriber(client)
    subscriber.subscribe(ScanStartMessage, scanner.run_scan)

    daemon.run()


def main():
    parser = CliParser()
    args = parser.parse_arguments()

    init_logging(
        "notus-scanner",
        args.log_level,
        log_file=args.log_file,
        foreground=args.foreground,
    )

    if not args.foreground:
        go_to_background()

    if not create_pid(args.pid_file):
        sys.exit()

    init_signal_handler(args.pid_file)

    logger.info("Starting notus-scanner version %s.", __version__)

    run_daemon(
        args.mqtt_broker_address,
        args.mqtt_broker_port,
        args.products_directory,
        args.disable_hashsum_verification,
    )


if __name__ == "__main__":
    main()



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, Optional, Union
from uuid import UUID, uuid4

from ..errors import MessageParsingError


class MessageType(Enum):
    RESULT = "result.scan"
    SCAN_STATUS = "scan.status"
    SCAN_START = "scan.start"


class Message:
    topic: str = None
    message_type: MessageType = None
    message_id: UUID = None
    group_id: str = None
    created: datetime = None

    def __init__(
        self,
        *,
        message_id: Optional[UUID] = None,
        group_id: Optional[str] = None,
        created: Optional[datetime] = None,
    ):
        self.message_id = message_id if message_id else uuid4()
        self.group_id = group_id if group_id else str(uuid4())
        self.created = created if created else datetime.utcnow()

    @classmethod
    def _parse(cls, data: Dict[str, Union[int, str]]) -> Dict[str, Any]:
        try:
            message_type = MessageType(data.get("message_type"))
        except ValueError as e:
            raise MessageParsingError(
                f"error while parsing 'message_type', {e}"
            ) from e
        if message_type != cls.message_type:
            raise MessageParsingError(
                f"Invalid message type {message_type} for {cls.__name__}. "
                f"Must be {cls.message_type}.",
            )

        try:
            message_id = UUID(data.get("message_id"))
        except (TypeError, ValueError) as e:
            raise MessageParsingError(
                f"error while parsing 'message_id', {e}"
            ) from e
        group_id = data.get("group_id")
        try:
            created = datetime.fromtimestamp(
                float(data.get("created")), timezone.utc
            )
        except (TypeError, ValueError) as e:
            raise MessageParsingError(
                f"error while parsing 'created', {e}"
            ) from e

        return {
            "message_id": message_id,
            "group_id": group_id,
            "created": created,
        }

    def serialize(self) -> Dict[str, Union[int, str]]:
        return {
            "message_id": str(self.message_id),
            "message_type": self.message_type.value
            if self.message_type
            else None,
            "group_id": str(self.group_id),
            "created": self.created.timestamp(),
        }

    @classmethod
    def deserialize(cls, data: Dict[str, Union[int, str]]) -> "Message":
        kwargs = cls._parse(data)
        return cls(**kwargs)

    @classmethod
    def load(cls, payload: Union[str, bytes]) -> "Message":
        data = json.loads(payload)
        return cls.deserialize(data)

    def dump(self) -> str:
        return json.dumps(self.serialize())

    def __str__(self) -> str:
        return self.dump()



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import argparse
import atexit
import logging
import signal
from functools import partial

from ..cli.parser import log_level
from ..config import DEFAULT_MQTT_BROKER_ADDRESS, DEFAULT_MQTT_BROKER_PORT
from ..messages.result import ResultMessage
from ..messages.status import ScanStatusMessage
from ..messaging.mqtt import MQTTClient, MQTTSubscriber


def print_scan_status(message: ScanStatusMessage):
    print("Scan Status")
    print("-----------")
    print(message.status)
    print("-----------\n")


def print_result_message(message: ResultMessage):
    print("Scan Result")
    print("-----------")
    print(f"OID {message.oid}")
    print(f"Host {message.host_ip} {message.host_name or ''}")
    print(message.value)
    print("-----------\n")


def disconnect(
    client: MQTTClient,
    _signum=None,
    _frame=None,
) -> None:
    client.disconnect()


def main():
    parser = argparse.ArgumentParser(
        description="A test client to show messages created by the Notus "
        "Scanner"
    )
    parser.add_argument(
        "-b",
        "--mqtt-broker-address",
        type=str,
        required=True,
        default=DEFAULT_MQTT_BROKER_ADDRESS,
        help="Hostname or IP address of the MQTT broker.",
    )
    parser.add_argument(
        "-p",
        "--mqtt-broker-port",
        type=int,
        default=DEFAULT_MQTT_BROKER_PORT,
        help="Port of the MQTT broker. (default: %(default)s)",
    )
    parser.add_argument(
        "-L",
        "--log-level",
        default="INFO",
        type=log_level,
        help="Wished level of logging (default: %(default)s)",
    )

    args = parser.parse_args()

    logging.basicConfig(level=args.log_level)

    client = MQTTClient(
        mqtt_broker_address=args.mqtt_broker_address,
        mqtt_broker_port=args.mqtt_broker_port,
        client_id="notus-subscriber",
    )
    client.connect()

    subscriber = MQTTSubscriber(client)
    subscriber.subscribe(ScanStatusMessage, print_scan_status)
    subscriber.subscribe(ResultMessage, print_result_message)

    disconnect_func = partial(disconnect, client)

    atexit.register(disconnect_func, client)
    signal.signal(signal.SIGTERM, disconnect_func)
    signal.signal(signal.SIGINT, disconnect_func)

    client.loop_forever()


if __name__ == "__main__":
    main()



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
from pathlib import Path
from typing import Any, Dict

from ..__version__ import __version__
from ..config import Config

ParserType = argparse.ArgumentParser
Arguments = argparse.Namespace

logger = logging.getLogger(__name__)

DEFAULT_CONFIG_FILE = "/etc/gvm/notus-scanner.toml"
DEFAULT_USER_CONFIG_FILE = "~/.config/notus-scanner.toml"


def log_level(string: str) -> str:
    """Check if provided string is a valid log level."""

    if not hasattr(logging, string.upper()):
        raise argparse.ArgumentTypeError(
            "log level must be one of {debug,info,warning,error,critical}"
        )
    return string.upper()


def _to_defaults(values: Dict[str, Any]) -> Dict[str, Any]:
    defaults = {}

    for key, value in values.items():
        defaults[key.replace("-", "_")] = value

    return defaults


class CliParser:
    def __init__(self) -> None:
        """Create a command-line arguments parser for Notus Scanner."""
        parser = argparse.ArgumentParser(
            description="Notus Scanner", add_help=False
        )

        parser.add_argument(
            "--version",
            help="Print version then exit.",
            action="version",
            version=f"%(prog)s {__version__}",
        )
        parser.add_argument(
            "-h",
            "--help",
            help="Show this help message and exit.",
            action="store_true",
        )

        parser.add_argument(
            "-c",
            "--config",
            nargs="?",
            help="Configuration file path. If not set %(prog)s "
            f"tries to load {DEFAULT_USER_CONFIG_FILE} and "
            f"{DEFAULT_CONFIG_FILE}.",
        )
        parser.add_argument(
            "--pid-file",
            help="Location of the file for the process ID "
            "(default: %(default)s)",
        )
        parser.add_argument(
            "-l",
            "--log-file",
            nargs="?",
            default=None,
            help="Log file path (default: syslog)",
        )
        parser.add_argument(
            "-L",
            "--log-level",
            type=log_level,
            help="Wished level of logging (default: %(default)s)",
        )
        parser.add_argument(
            "-f",
            "--foreground",
            action="store_true",
            help="Run in foreground and logs all messages to console.",
        )
        parser.add_argument(
            "--products-directory",
            type=Path,
            help=(
                "Choose a custom directory that contains product advisory "
                "files generated by the Notus Generator. (default: %(default)s)"
            ),
        )
        parser.add_argument(
            "-b",
            "--mqtt-broker-address",
            type=str,
            help="Hostname or IP address of the MQTT broker. "
            "(default: %(default)s)",
        )
        parser.add_argument(
            "-p",
            "--mqtt-broker-port",
            type=int,
            help="Port of the MQTT broker. (default: %(default)s)",
        )
        parser.add_argument(
            "--disable-hashsum-verification",
            type=bool,
            default=False,
            help="Disables hashsum verification (default: %(default)s)",
        )
        self.parser = parser

    def _set_defaults(self, configfilename=None) -> None:
        config_data = self._load_config(configfilename)
        self.parser.set_defaults(**_to_defaults(config_data))

    def _load_config(self, configfile: str) -> Dict[str, Any]:
        config = Config()

        use_default = configfile is None
        configpath = None

        if use_default:
            for file in [DEFAULT_USER_CONFIG_FILE, DEFAULT_CONFIG_FILE]:
                path = Path(file).expanduser().resolve()
                if path.exists():
                    configpath = path
                    break
                else:
                    logger.debug(
                        "Ignoring non existing config file %s", configfile
                    )

            if not configpath:
                return config.values()
        else:
            configpath = Path(configfile).expanduser().resolve()
            if not configpath.exists():
                logger.warning(
                    "Ignoring non existing config file %s", configfile
                )
                return config.values()

        config.load(configpath)
        logger.debug("Loaded config %s", configfile)

        return config.values()

    def parse_arguments(self, args=None) -> Arguments:
        # Parse args to get the config file path passed as option
        known_args, _ = self.parser.parse_known_args(args)

        # Load the defaults from the config file if it exists.
        # This also override what was passed as cmd option.
        self._set_defaults(known_args.config)

        if known_args.help:
            self.parser.print_help()
            self.parser.exit(0)

        return self.parser.parse_args(args)



# Copyright (C) 2014-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import atexit
import logging
import os
import signal
import sys
from functools import partial
from logging.handlers import SysLogHandler, WatchedFileHandler
from pathlib import Path
from typing import Optional, Union

import psutil

from .__version__ import __version__

logger = logging.getLogger(__name__)


def go_to_background() -> None:
    """Daemonize the running process."""
    try:
        if os.fork():
            sys.exit()
    except OSError as errmsg:
        logger.error("Fork failed: %s", errmsg)
        sys.exit(1)


def create_pid(pid_file: str) -> bool:
    """Check if there is an already running daemon and creates the pid file.
    Otherwise gives an error."""

    pid = os.getpid()
    new_process = psutil.Process(pid)
    new_process_name = new_process.name()

    pid_path = Path(pid_file)

    if pid_path.is_file():
        process_name = None
        current_pid = pid_path.read_text(encoding="utf-8").strip()
        try:
            current_pid = int(current_pid)
        except (TypeError, ValueError):
            current_pid = None

        if current_pid:
            try:
                process = psutil.Process(current_pid)
                process_name = process.name()
            except psutil.NoSuchProcess:
                pass

            if process_name == new_process_name:
                logger.error(
                    "There is an already running process. See %s.",
                    str(pid_path.absolute()),
                )
                return False
            else:
                logger.debug(
                    "There is an existing pid file '%s', but the PID %s belongs"
                    " to the process %s. It seems that %s was abruptly stopped."
                    " Removing the pid file.",
                    str(pid_path.absolute()),
                    current_pid,
                    process_name,
                    new_process_name,
                )

    try:
        pid_path.write_text(str(pid), encoding="utf-8")
    except (FileNotFoundError, PermissionError) as e:
        logger.error(
            "Failed to create pid file %s. %s", str(pid_path.absolute()), e
        )
        return False

    return True


def exit_cleanup(
    pid_file: str,
    _signum=None,
    _frame=None,
) -> None:
    """Removes the pid_file before ending the daemon."""
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    pid_path = Path(pid_file)

    if not pid_path.is_file():
        return

    with pid_path.open(encoding="utf-8") as f:
        if int(f.read()) == os.getpid():
            logger.debug("Finishing daemon process")
            pid_path.unlink()
            sys.exit()


def init_signal_handler(pid_file: str):
    atexit.register(exit_cleanup, pid_file=pid_file)
    signal.signal(signal.SIGTERM, partial(exit_cleanup, pid_file))
    signal.signal(signal.SIGINT, partial(exit_cleanup, pid_file))


def init_logging(
    name: str,
    log_level: Union[int, str],
    *,
    log_file: Optional[str] = None,
    foreground: Optional[bool] = False,
):

    rootlogger = logging.getLogger()
    rootlogger.setLevel(log_level)

    if foreground:
        console = logging.StreamHandler()
        console.setFormatter(
            logging.Formatter(
                f"%(asctime)s {name}: %(levelname)s: (%(name)s) %(message)s"
            )
        )
        rootlogger.addHandler(console)
    elif log_file:
        logfile = WatchedFileHandler(log_file)
        logfile.setFormatter(
            logging.Formatter(
                f"%(asctime)s {name}: %(levelname)s: (%(name)s) %(message)s"
            )
        )
        rootlogger.addHandler(logfile)
    else:
        syslog = SysLogHandler("/dev/log")
        syslog.setFormatter(
            logging.Formatter(f"{name}: %(levelname)s: (%(name)s) %(message)s")
        )
        rootlogger.addHandler(syslog)

    return rootlogger



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from enum import Enum
from typing import Any, Dict, Optional, Union
from uuid import UUID

from .message import Message, MessageType


class ScanStatus(Enum):
    FINISHED = "finished"
    REQUESTED = "requested"
    QUEUED = "queued"
    INIT = "init"
    RUNNING = "running"
    STOPPING = "stopping"
    STOPPED = "stopped"
    INTERRUPTED = "interrupted"


class ScanStatusMessage(Message):
    message_type = MessageType.SCAN_STATUS
    topic = "scanner/status"

    def __init__(
        self,
        *,
        scan_id: str,
        host_ip: str,
        status: ScanStatus,
        message_id: Optional[UUID] = None,
        group_id: Optional[str] = None,
        created: Optional[datetime] = None,
    ):
        super().__init__(
            message_id=message_id, group_id=group_id, created=created
        )
        self.scan_id = scan_id
        self.host_ip = host_ip
        self.status = status

    def serialize(self) -> Dict[str, Union[int, str]]:
        message = super().serialize()
        message.update(
            {
                "scan_id": self.scan_id,
                "host_ip": self.host_ip,
                "status": self.status.value,
            }
        )
        return message

    @classmethod
    def _parse(cls, data: Dict[str, Union[int, str]]) -> Dict[str, Any]:
        kwargs = super()._parse(data)
        kwargs.update(
            {
                "scan_id": data.get("scan_id"),
                "host_ip": data.get("host_ip"),
                "status": ScanStatus(data.get("status")),
            }
        )
        return kwargs



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Module for parsing and comparing Debian packages (.deb)
"""

import logging
import re
from dataclasses import dataclass

from .package import Package, PackageComparison

_deb_compile = re.compile(r"(.*)-(?:(\d*):)?(\d.*)-(.*)")
_deb_compile_wo_revision = re.compile(r"(.*)-(?:(\d*):)?(\d.*)")
_deb_compile_version = re.compile(r"(?:(\d*):)?(\d.*)-(.*)")
_deb_compile_version_wo_revision = re.compile(r"(?:(\d*):)?(\d.*)")


logger = logging.getLogger(__name__)


@dataclass
class DEBPackage(Package):
    """Represents a .deb based package"""

    upstream_version: str
    debian_revision: str
    epoch: str

    __hash__ = Package.__hash__

    def compare(self, other: "DEBPackage") -> PackageComparison:
        if self.name != other.name:
            return PackageComparison.NOT_COMPARABLE

        if self.full_version == other.full_version:
            return PackageComparison.EQUAL

        if self.epoch != other.epoch:
            return (
                PackageComparison.A_NEWER
                if self.epoch > other.epoch
                else PackageComparison.B_NEWER
            )

        comp = self.version_compare(
            self.upstream_version, other.upstream_version
        )
        if comp != PackageComparison.EQUAL:
            return comp

        return self.version_compare(self.debian_revision, other.debian_revision)

    @staticmethod
    def from_full_name(full_name: str):
        if not full_name:
            return None

        full_name = full_name.strip()

        match = _deb_compile.match(full_name)
        # Try to get data with
        if match:
            name, epoch, upstream_version, debian_revision = match.groups()
        else:
            match = _deb_compile_wo_revision.match(full_name)
            if match:
                name, epoch, upstream_version = match.groups()
                debian_revision = ""
            else:
                logger.warning(
                    "The deb package %s could not be parsed", full_name
                )
                return None

        if not epoch:
            epoch = "0"
            full_version = f"{upstream_version}"
        else:
            full_version = f"{epoch}:{upstream_version}"

        if debian_revision:
            full_version = f"{full_version}-{debian_revision}"

        return DEBPackage(
            name=name,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
            epoch=epoch,
            upstream_version=upstream_version,
            debian_revision=debian_revision,
        )

    @staticmethod
    def from_name_and_full_version(name: str, full_version: str):
        if not name or not full_version:
            return None
        name = name.strip()
        full_version = full_version.strip()
        match = _deb_compile_version.match(full_version)
        if match:
            (
                epoch,
                upstream_version,
                debian_revision,
            ) = match.groups()
        else:
            match = _deb_compile_version_wo_revision.match(full_version)
            if match:
                (
                    epoch,
                    upstream_version,
                ) = match.groups()
                debian_revision = ""
            else:
                logger.warning(
                    "The deb package %s %s could not be parsed",
                    name,
                    full_version,
                )
                return None

        if not epoch:
            epoch = "0"

        return DEBPackage(
            name=name,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
            epoch=epoch,
            upstream_version=upstream_version,
            debian_revision=debian_revision,
        )



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Test module for command line arguments.
"""

import tempfile
import unittest
from pathlib import Path
from typing import List

from notus.scanner.cli.parser import Arguments, CliParser
from notus.scanner.config import (
    DEFAULT_LOG_LEVEL,
    DEFAULT_MQTT_BROKER_ADDRESS,
    DEFAULT_MQTT_BROKER_PORT,
    DEFAULT_PID_FILE,
    DEFAULT_PRODUCTS_DIRECTORY,
)


class CliParserTestCase(unittest.TestCase):
    def setUp(self):
        self.parser = CliParser()

    def parse_args(self, args: List[str]) -> Arguments:
        return self.parser.parse_arguments(args)

    def test_mqtt_broker(self):
        args = self.parse_args(["--mqtt-broker-address=localhost"])
        self.assertEqual("localhost", args.mqtt_broker_address)

        args = self.parse_args(["-b", "localhost"])
        self.assertEqual("localhost", args.mqtt_broker_address)

    def test_mqtt_broker_port(self):
        args = self.parse_args(["--mqtt-broker-port=12345"])
        self.assertEqual(args.mqtt_broker_port, 12345)

        args = self.parse_args(["-p", "12345"])
        self.assertEqual(args.mqtt_broker_port, 12345)

    def test_correct_upper_case_log_level(self):
        args = self.parse_args(["--log-level=ERROR"])
        self.assertEqual("ERROR", args.log_level)

    def test_correct_lower_case_log_level(self):
        args = self.parse_args(["-L", "info"])
        self.assertEqual("INFO", args.log_level)

    def test_advisories_directory(self):
        args = self.parse_args(["--products-directory=/tmp"])
        self.assertEqual(Path("/tmp"), args.products_directory)

    def test_pid_file(self):
        args = self.parse_args(["--pid-file=/foo/bar"])
        self.assertEqual(args.pid_file, "/foo/bar")

    def test_log_file(self):
        args = self.parse_args(["--log-file=/foo/bar"])
        self.assertEqual(args.log_file, "/foo/bar")

        args = self.parse_args(["-l", "/foo/bar"])
        self.assertEqual(args.log_file, "/foo/bar")

    def test_foreground(self):
        args = self.parse_args(["--foreground"])
        self.assertTrue(args.foreground)

        args = self.parse_args(["-f"])
        self.assertTrue(args.foreground)

    def test_disable_hashsum_verification(self):
        args = self.parse_args(["--disable-hashsum-verification=true"])
        self.assertTrue(args.disable_hashsum_verification)

    def test_defaults(self):
        args = self.parse_args([])

        self.assertEqual(
            args.products_directory, Path(DEFAULT_PRODUCTS_DIRECTORY)
        )
        self.assertIsNone(args.config)
        self.assertIsNone(args.log_file)
        self.assertEqual(args.log_level, DEFAULT_LOG_LEVEL)
        self.assertEqual(args.mqtt_broker_port, DEFAULT_MQTT_BROKER_PORT)
        self.assertEqual(args.mqtt_broker_address, DEFAULT_MQTT_BROKER_ADDRESS)
        self.assertEqual(args.pid_file, DEFAULT_PID_FILE)
        self.assertEqual(args.disable_hashsum_verification, False)
        self.assertFalse(args.foreground)

    def test_config_file_provide_mqtt_broker_address(self):
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(b"[notus-scanner]\nmqtt-broker-address='1.2.3.4'")
            fp.flush()

            args = self.parse_args(["-c", fp.name])
            self.assertEqual(args.mqtt_broker_address, "1.2.3.4")

    def test_config_file(self):
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(
                b"""[notus-scanner]
                mqtt-broker-address="1.2.3.4"
                mqtt-broker-port="123"
                products-directory="/tmp"
                pid-file="foo.bar"
                log-file="foo.log"
                log-level="DEBUG"
                """
            )
            fp.flush()

            args = self.parse_args(["-c", fp.name])

            self.assertEqual(args.mqtt_broker_address, "1.2.3.4")
            self.assertEqual(args.mqtt_broker_port, 123)
            self.assertEqual(args.products_directory, Path("/tmp"))
            self.assertEqual(args.pid_file, "foo.bar")
            self.assertEqual(args.log_file, "foo.log")
            self.assertEqual(args.log_level, "DEBUG")



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from typing import Dict, Optional
from unittest import TestCase
from unittest.mock import Mock, patch

from notus.scanner.loader.gpg_sha_verifier import (
    ReloadConfiguration,
    VerificationResult,
    create_verify,
    gpg_sha256sums,
    reload_sha256sums,
)


class GpgTest(TestCase):
    @patch("gnupg.GPG")
    @patch("pathlib.Path")
    def test_reload(self, gmock, pathmock: Path):
        def on_failure(_: Optional[Dict[str, str]]) -> Dict[str, str]:
            raise Exception("verification_failed")

        omock = Mock()
        emock = Mock()
        omock.__enter__ = Mock(return_value=emock)
        omock.__exit__ = Mock()
        pathmock.open.return_value = omock
        emock.readlines.side_effect = [["h  hi\n"], ["g  gude\n"]]
        emock.read.side_effect = [b"hi", b"", b"hi", b"", b"ih", b""]

        load = reload_sha256sums(
            ReloadConfiguration(
                hash_file=pathmock,
                on_verification_failure=on_failure,
                gpg=gmock,
            )
        )
        self.assertDictEqual(load(), {"h": "hi"})
        self.assertDictEqual(load(), {"h": "hi"})
        self.assertDictEqual(load(), {"g": "gude"})
        gmock.verify_file.side_effect = [False]
        with self.assertRaises(Exception):
            load()

    @patch("gnupg.GPG")
    @patch("pathlib.Path")
    def test_verifying(self, gmock, pathmock: Path):
        omock = Mock()
        emock = Mock()
        omock.__enter__ = Mock(return_value=emock)
        omock.__exit__ = Mock()
        pathmock.open.return_value = omock
        emock.readlines.side_effect = [["h  hi\n", "g  gude\n"]]
        success_result = gpg_sha256sums(pathmock, gmock)
        self.assertIsNotNone(success_result)
        self.assertDictEqual(success_result, {"h": "hi", "g": "gude"})
        gmock.verify_file.side_effect = [False]
        self.assertIsNone(gpg_sha256sums(pathmock, gmock))

    @patch("pathlib.Path")
    def test_verify_closure(self, pathmock):
        shas = (
            "98ea6e4f216f2fb4b69fff9b3a44842c38686ca685f3f55dc48c5d3fb1107be4"
        )
        vsuccess = create_verify(lambda: {shas: "hi.txt"})
        omock = Mock()
        emock = Mock()
        omock.__enter__ = Mock(return_value=emock)
        omock.__exit__ = Mock()
        pathmock.open.return_value = omock
        emock.read.side_effect = [bytes("hi\n", "utf-8"), ""]
        pathmock.name = "hi.txt"
        self.assertEqual(vsuccess(pathmock), VerificationResult.SUCCESS)
        emock.read.side_effect = [bytes("hi\n", "utf-8"), ""]
        pathmock.name = "false.txt"
        self.assertEqual(vsuccess(pathmock), VerificationResult.INVALID_NAME)
        emock.read.side_effect = [bytes("hin", "utf-8"), ""]
        pathmock.name = "hi.txt"
        self.assertEqual(vsuccess(pathmock), VerificationResult.INVALID_HASH)



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Module for parsing and comparing Gentoo packages (ebuild)

Currently the ebuild are based upon find within `/usr/portage` and are
manipulated as:
sed 's,/usr/portage/([^/]*)/.*/([^/]*).ebuild$,\1/\2,'

Meaning the corresponding name will be:
x11-plugins/wmix-3.5

The version generated by our generator will be delivered via name and version.

The version of the program is program specific hence we could just use Version
of pep440 to guess immediately.
"""

import logging
from dataclasses import dataclass

from .package import Package, PackageComparison

logger = logging.getLogger(__name__)


@dataclass(eq=False)
class EBuildPackage(Package):
    """Represents a .ebuild based package"""

    def compare(self, other: Package) -> PackageComparison:
        if self.name != other.name:
            return PackageComparison.NOT_COMPARABLE

        if self.full_version == other.full_version:
            return PackageComparison.EQUAL

        return self.version_compare(self.full_version, other.full_version)

    @staticmethod
    def from_full_name(full_name: str):
        if not full_name:
            return None
        base_name = (
            full_name[full_name.index("/") + 1 :]
            if "/" in full_name
            else full_name
        )
        if not base_name:
            return None
        d_index = base_name.index("-") if "-" in base_name else len(base_name)
        full_version = base_name[d_index + 1 :]
        if not full_version:
            return None
        return EBuildPackage(
            name=full_name[: len(full_name) - len(base_name[d_index:])],
            full_name=full_name,
            full_version=full_version,
        )

    @staticmethod
    def from_name_and_full_version(name: str, full_version: str):
        if not name or not full_version:
            return None
        name = name.strip()
        full_version = full_version.strip()
        return EBuildPackage(
            name=name,
            full_name=f"{name}-{full_version}",
            full_version=full_version,
        )
